#include "fmus.h"
#include "text.h"
#include "fmus_io.h"
#include "mpfont.h"
#include "cli.h"

#include <math.h>

#define GCLIP_FLT

// [ ] kerning
// [ ] change unknown 1 on multiline text and see if anything happens
// [ ] use freetype to pull line height and set that correctly?
// [ ] what happens if you override the texture format?
// [x] -g edit 0x41 lpad=128 should fail
// [x] dont' allow extra arguments after command line with -b, -e, (-q?)
// [x] "./fp -w -b -g add 0x" does not print error (doesn't add glyph tho)

#ifdef FMUS_DBGMEM
static u32_t mallocs=0;
static u32_t frees=0;
#endif

// GCLIP compatibility
#define GCLIP_MALLOCMSG      FMUS_MALLOCMSG
#define GCLIP_REALLOCMSG     FMUS_REALLOCMSG
#define GCLIP_FREEMSG        FMUS_FREEMSG
#define GCLIP_ESTREAM        stdout
#define gclip_scprintf_dummy fmus_scprintf_dummy
#define GCLIP_E_NULL         FMUS_E_NULL
#define GCLIP_E_OK           FMUS_E_OK
#define GCLIP_E_MALLOC       FMUS_E_MALLOC
#define GCLIP_SCPRINTF_1ARG  FMUS_SCPRINTF_1ARG
#define GCLIP_E_BUG          FMUS_E_BUG
#define GCLIP_E_BADINDENT    1000
#define GCLIP_FMT_SIZET      FMUS_FMT_SIZET
#define GCLIP_E_ZEROWIDTH    1001
#define GCLIP_64
#define GCLIP_FMT_S64        FMUS_FMT_S64
#define GCLIP_FMT_U64        FMUS_FMT_U64
#define GCLIP_FMT_U64HEX     FMUS_FMT_U64HEX

char fmus_scprintf_dummy;

#define SHEET_MAX_SIZE 1024 // some JP fonts go up to > 640 px, so
#define SHEET_MIN_SIZE 32

static fmus_err_t init_fmus(fmus_t *fmus);
static void finish_fmus(fmus_t *fmus);
static s32_t fmus_cli_get_term_width (void);
static fmus_err_t fmus_draw_bitmaps (fmus_t *fmus,
                                     fmus_mpfont_t *mpfont,
                                     s32_t face_ix,
                                     float px_size,
                                     u8_t ***bitmaps);
static u8_t test_pixel_block (s32_t glyph_w,
                              s32_t glyph_h,
                              s32_t test_x,
                              s32_t test_y,
                              s32_t test_sheet_size,
                              u8_t curr_sheet_L2[SHEET_MAX_SIZE][SHEET_MAX_SIZE]);
static s32_t find_max_occupied_x_coord (u8_t curr_sheet_L2[SHEET_MAX_SIZE][SHEET_MAX_SIZE],
                                        s32_t test_sheet_size);
static s32_t find_max_occupied_y_coord (u8_t curr_sheet_L2[SHEET_MAX_SIZE][SHEET_MAX_SIZE],
                                        s32_t test_sheet_size);
static fmus_err_t place_glyphs (fmus_t *fmus,
                                u32_t tx_fmt,
                                s32_t test_sheet_size,
                                u8_t (*sheets)[3][SHEET_MAX_SIZE][SHEET_MAX_SIZE],
                                u32_t *glyph_ix_inout,
                                u8_t *current_sheet_inout);
static fmus_err_t export_texture(fmus_t *fmus,
                                 u32_t tx_fmt,
                                 s32_t sheet_w,
                                 s32_t sheet_h,
                                 u8_t sheets[4][3][SHEET_MAX_SIZE][SHEET_MAX_SIZE],
                                 u8_t **txtr_out,
                                 size_t *txtr_len_out);
static gclip_err_t gclip_slice_printf_grow(char **buf,
                                           size_t *alloced);
static gclip_err_t gclip_free_lineslice(char *buf);
static gclip_err_t gclip_lineslice(u16_t cols,
                                   char *in,
                                   char **out,
                                   size_t *cursor_pos, /* both input and output */
                                   size_t wrap_indent);
static fmus_err_t try_sheet_sizes(fmus_t *fmus,
                                  u8_t (*sheets)[3][SHEET_MAX_SIZE][SHEET_MAX_SIZE],
                                  s32_t *sheet_size_out);
static fmus_err_t apply_pre_render_transformations (fmus_t *fmus);
static fmus_err_t apply_post_render_transformations (fmus_t *fmus);
static fmus_err_t fmus_main_do_txtr (fmus_t *fmus, fmus_mpfont_t *mpfont);
static fmus_err_t fmus_main (fmus_t *fmus);
static void update_kerning_indices (fmus_t *fmus, fmus_mpfont_t *mpfont);
static void adjust_glyph_spacing(fmus_glyph_t *glyph, s8_t adjust);

#ifdef FMUS_WINDOZE // stupid visual studio
float roundf(float f) {
  return floorf(f + 0.5f);
}

char *fmus_windoze_strerror(void) {
  return strerror(errno); // FIXME
}
#endif

static void update_kerning_indices (fmus_t *fmus, fmus_mpfont_t *mpfont) {
  u32_t i;
  u32_t kerning_entries_new;
  kerning_entries_new=0;
  // phase 1: remove adjust=0 entries from kerning table
  for (i=0;i<mpfont->kerning_entries;i++) {
    if (fmus->kerning[i].adjust != 0) {
      fmus->kerning[kerning_entries_new] = fmus->kerning[i];
      kerning_entries_new++;
    }
  }
  if (kerning_entries_new != mpfont->kerning_entries)
    FMUS_MPRINTF(fmus->tw, "Compacted kerning table: %u entries -> %u entries.\n", 
                 mpfont->kerning_entries, kerning_entries_new);
  mpfont->kerning_entries = kerning_entries_new;
  // phase 2: update all glyphs
  for (i=0;i<mpfont->num_glyphs;i++) {
    u32_t j;
    for (j=0;j<mpfont->kerning_entries;j++)
      if (fmus->glyphs[i].utf16 == fmus->kerning[j].c1_utf16)
        break;
    fmus->glyphs[i].kern_ix = j;
  }
}

static fmus_err_t apply_pre_render_transformations (fmus_t *fmus) {
  
  u32_t i;
  //size_t txid_len;
  //char *txid_s;
  fmus_err_t e;
  
  e = FMUS_E_OK;
  
  if (fmus->font_px_size >= 0.0) {
    fmus->mpfont.font_size = fmus->font_px_size;
    FMUS_MPRINTF(fmus->tw, "Using user-provided font size of %f.\n", fmus->font_px_size);
  } else {
    if (!fmus->edit_mode && !fmus->do_not_load)
      FMUS_MPRINTF(fmus->tw, "Using font size of %u extracted from source FONT file.\n",
                   fmus->mpfont.font_size); 
  }
  
  if (fmus->override_op_font_ver) {
    FMUS_MPRINTF(fmus->tw, "Overriding font version to %u.\n", fmus->op_font_ver);
    fmus->mpfont.version = fmus->op_font_ver;
  }
  if (fmus->override_op_tx_fmt) {
    FMUS_MPRINTF(fmus->tw, "Overriding texture format to %u.\n", fmus->op_tx_fmt);
    fmus->mpfont.tx_fmt = fmus->op_tx_fmt;
  }
  if (fmus->override_op_font_name != NULL) {
    FMUS_MPRINTF(fmus->tw, "Overriding output font name to \"%s\".\n", fmus->override_op_font_name);
    FMUS_FREEMSG(fmus->mpfont.name, "mpfont->name");
    fmus->mpfont.name = FMUS_MALLOCMSG(strlen(fmus->override_op_font_name) + 1, "mpfont->name");
    memcpy (fmus->mpfont.name, fmus->override_op_font_name, 1 + strlen(fmus->override_op_font_name));
  }
    
  // do glyphmods
  for (i=0;i<fmus->glyphmods_fill;i++) {
    // see if we already have a glyph for this glyphmod
    u32_t j, k;
    fmus_glyphmod_t *gm;
    fmus_glyph_t *glyph;
    glyph=NULL;
    gm = fmus->glyphmods + i;
    switch (gm->mod_type) {
      case GLYPHMOD_ADD:
      case GLYPHMOD_DELETE:
      case GLYPHMOD_EDIT:
        break;
      default:
        FMUS_EPRINTF(fmus->tw, "Error: BUG: apply_pre_render_transformations(): illegal gm->mod_type(%u)\n", gm->mod_type);
        return FMUS_E_BUG;
    }
//printf("gm.glyph.utf16=x%x, mod_type=%u, lpad=%u\n", gm->glyph.utf16, gm->mod_type, gm->glyph.lpad);
//printf("num_glyphs = %u\n", fmus->mpfont.num_glyphs);
    for (j=0;j<fmus->mpfont.num_glyphs;j++) {
      if (gm->glyph.utf16 == fmus->glyphs[j].utf16) {
        glyph = fmus->glyphs + j;
        break;
      }
    }
    if (glyph) { // must be an existing glyph for edit/delete, but none for add
      if (gm->mod_type == GLYPHMOD_ADD) {
        FMUS_EPRINTF(fmus->tw, "Error: Could not add glyph with Unicode 0x%x -- "
                               "that glyph already exists in the input FONT!\n", gm->glyph.utf16);
        return FMUS_E_CLI;
      }
    } else {
      if (gm->mod_type != GLYPHMOD_ADD) {
        FMUS_EPRINTF(fmus->tw, "Could not edit or delete glyph with Unicode 0x%x -- that "
                               "glyph does not exist in the input FONT!\n", gm->glyph.utf16);
        return FMUS_E_CLI;
      }
    }
    
    if (gm->mod_type == GLYPHMOD_ADD) {
      
      (fmus->mpfont.num_glyphs)++;
//printf("allocing glyphs[1]\n");
      fmus->glyphs = FMUS_REALLOCMSG(fmus->glyphs, sizeof(fmus_glyph_t) * (fmus->mpfont.num_glyphs),
                                     "fmus->glyphs");
      memset(fmus->glyphs + (fmus->mpfont.num_glyphs-1),
             0,
             sizeof(fmus_glyph_t));
      fmus->glyphs[fmus->mpfont.num_glyphs-1].utf16 = gm->glyph.utf16;
      
    } else if (gm->mod_type == GLYPHMOD_DELETE) {
      
      (fmus->mpfont.num_glyphs)--;
      memmove(glyph,
              glyph + 1,
              (fmus->mpfont.num_glyphs - j) * sizeof(fmus_glyph_t));  
                          
    } else { // GLYPHMOD_EDIT
      
      for (k=0;k<NUM_GLYPH_EDIT_LABELS;k++) {
        if (gm->gelixes[k]) {
          switch (k) {
            case GELIX_TXCL:    glyph->tx_left = gm->glyph.tx_left; break;
            case GELIX_TXCT:    glyph->tx_top = gm->glyph.tx_top; break;
            case GELIX_TXCR:    glyph->tx_right = gm->glyph.tx_right; break;
            case GELIX_TXCB:    glyph->tx_bottom = gm->glyph.tx_bottom; break;
            case GELIX_LSB:     glyph->lsb = gm->glyph.lsb; break;
            case GELIX_LWIDTH:  glyph->layout_width = gm->glyph.layout_width; break;
            case GELIX_RSB:     glyph->rsb = gm->glyph.rsb; break;
            case GELIX_DWIDTH:  glyph->display_width = gm->glyph.display_width; break;
            case GELIX_HEIGHT:  glyph->height = gm->glyph.height; break;
            case GELIX_VERTOFF: glyph->vert_offset = gm->glyph.vert_offset; break;
            default: break; // FIXME: bug
          }
        }
      } // next gelix
    } // endif edit mode
  } // next glyphmod
  
  if (fmus->override_line_spacing)
    fmus->mpfont.line_height = fmus->line_spacing;
  
  if (fmus->do_y_stretch) {
    for (i=0;i<fmus->mpfont.num_glyphs;i++) {
      fmus_glyph_t *glyph;
      glyph = fmus->glyphs + i;
      glyph->height = (u32_t) roundf(((float) fmus->glyphs[i].height) * fmus->y_stretch);
      glyph->vert_offset = 0xff & ((u32_t) roundf(((float) glyph->vert_offset) * fmus->y_stretch));
    }
  }
  
  if (fmus->do_lgap_adjust)
    for (i=0;i<fmus->mpfont.num_glyphs;i++)
      adjust_glyph_spacing(fmus->glyphs + i, fmus->lgap_adjust);
  
  if (fmus->do_wgap_adjust)
    for (i=0;i<fmus->mpfont.num_glyphs;i++)
      if (fmus->glyphs[i].utf16 == ' ')
        adjust_glyph_spacing(fmus->glyphs + i, fmus->wgap_adjust);
        
  if (fmus->omit_kerning) {
    FMUS_FREEMSG(fmus->kerning, "fmus->kerning");
    fmus->kerning = NULL;
    fmus->mpfont.kerning_entries = 0;
  }
  
  if (fmus->txid_string != NULL) {
    // override texture ID
    // the string from the command line
    // has not been validated yet
    e = cli_parse_txid (fmus, fmus->mpfont.version, fmus->txid_string, &(fmus->mpfont.txid_64));
    if (e != FMUS_E_OK)
      return e;
  }
  
  return FMUS_E_OK;
  
}

static fmus_err_t apply_post_render_transformations(fmus_t *fmus) {
  size_t watermark_len;
  char *watermark_start;
  size_t namelen;
  char *name_new;
  // search for previous watermark
  watermark_len = strlen(WATERMARK);
  watermark_start = strstr ((char *) fmus->mpfont.name, (char *) WATERMARK_1);
  namelen = strlen(fmus->mpfont.name);
  
  if (!fmus->edit_mode) { // don't do watermarking in edit mode
    if ((NULL != watermark_start) && (NULL == fmus->override_op_font_name)) {
      // deal with existing watermark
      if (!fmus->do_not_load) { // don't display the warning in blank-font mode
        FMUS_WPRINTF(fmus->tw, "Warning: This .FONT file was previously built "
                               "by Fontimus Prime. You are basing your new .FONT "
                               "on an already-modified .FONT -- are you sure "
                               "you want to do this?");
      }
      name_new = FMUS_MALLOCMSG(namelen + watermark_len + 1, "mpfont->name");
      memcpy (name_new, fmus->mpfont.name, watermark_start - fmus->mpfont.name);
      memcpy (name_new + (watermark_start - fmus->mpfont.name), WATERMARK, watermark_len + 1); 
      FMUS_FREEMSG(fmus->mpfont.name, "mpfont->name");
      fmus->mpfont.name = name_new;
    } else if (NULL == fmus->override_op_font_name) {
      // append new watermark
      name_new = FMUS_MALLOCMSG(namelen + watermark_len + 1, "mpfont->name");
      memcpy (name_new, fmus->mpfont.name, namelen);
      memcpy (name_new + namelen, WATERMARK, watermark_len + 1); 
      FMUS_FREEMSG(fmus->mpfont.name, "mpfont->name");
      fmus->mpfont.name = name_new;
    }
  }
  return FMUS_E_OK;
}

static void adjust_glyph_spacing(fmus_glyph_t *glyph, s8_t adjust) {
  
  s32_t adjust_lsb;
  s32_t adjust_rsb;
  s32_t lsb;
  s32_t rsb;
  //u32_t remainder=0;
  
  adjust_lsb = adjust / 2;
  adjust_rsb = adjust - adjust_lsb;
  lsb = glyph->lsb;
  rsb = glyph->rsb;
  lsb += adjust_lsb;
  rsb += adjust_rsb;
  
  if (adjust >= 0) {
    if (lsb > 127)
      glyph->lsb = 127;
    else
      glyph->lsb = lsb;
    if (rsb > 127)
      glyph->rsb = 127;
    else
      glyph->rsb = rsb;
  } else {
    // adjust < 0 ...
    if (lsb < -128)
      glyph->lsb = -128;
    else
      glyph->lsb = lsb;
    if (rsb < -128)
      glyph->rsb = -128;
    else
      glyph->rsb = rsb;
    /*
    if (lpad < 0) {
      remainder = -lpad;
      glyph->lpad = 0;
    }
    if (rpad < 0) {
      remainder += -rpad;
      glyph->rpad = 0;
    }
    if (remainder > 0) {
      // try lpad again
      lpad = glyph->lpad;
      lpad -= remainder;
      if (lpad < 0) {
        remainder = -lpad;
        glyph->lpad = 0;
      }
    }
    if (remainder > 0) {
      // take anything remaining off the layout width
      if (remainder > glyph->layout_width)
        glyph->layout_width = 0;
      else
        (glyph->layout_width) -= remainder;
    }
    */
  }
}

#ifdef FMUS_WINDOZE
int wmain(int argc, wchar_t *argv[], wchar_t *envp[]) {
#else
int main (int argc, char *argv[]) {
#endif

  s32_t tmpcols;
  fmus_err_t e;
  fmus_t fmus;
  char **my_argv = NULL;
  int j;

  init_fmus(&fmus); // memset(0)
  
  // convert argv[] from wide char to multibyte unicode
  my_argv = (char **) FMUS_MALLOCMSG(sizeof(char *) * argc, "multibyte argv[1]");
  
  for (j=0;j<argc;j++) {
    
	  size_t size=0;
	  
	  fmus_scprintf_dummy = 0;

#ifdef FMUS_WINDOZE
	  size = WideCharToMultiByte (CP_UTF8, 0, argv[j], -1, NULL, 0, NULL, NULL);
#else
    size = 1 + strlen(argv[j]);
#endif
	  my_argv[j] = (char *) FMUS_MALLOCMSG(size, "multibyte argv[2]");
#ifdef FMUS_WINDOZE
	  WideCharToMultiByte (CP_UTF8, 0, argv[j], -1, my_argv[j], size, NULL, NULL);
#else
    memcpy(my_argv[j], argv[j], size);
#endif

  }
  
  if (-1 != (tmpcols = fmus_cli_get_term_width()))
    fmus.tw = 0xffff & tmpcols;
  else
    fmus.tw = 80; /* otherwise use default 80 */

  e = parse_cli (&fmus, argc, my_argv);
  if (e != FMUS_E_OK)
    return e;
  
  e = fmus_main(&fmus);

  for (j=0;j<argc;j++)
    FMUS_FREEMSG(my_argv[j], "multibyte argv[2]");
  FMUS_FREEMSG(my_argv, "multibyte argv[1]");

  finish_fmus(&fmus);

#if FMUS_DBGMEM
  printf("allocs %u, frees %u.\n", fmus_dbgmem_get_allocs(), fmus_dbgmem_get_frees());
  if (fmus_dbgmem_get_allocs() != fmus_dbgmem_get_frees())
    e = FMUS_E_ALLOCS_FREES;
#endif

  return e;
  
}

static fmus_err_t fmus_main (fmus_t *fmus) {

  u8_t *newfont;
  size_t newfont_len;
  fmus_err_t e;

  init_mpfont(&(fmus->mpfont));
  
  printf("\n");
  
  if (!fmus->do_not_load) { // not -b
    e = fmus_load_mpfont (fmus, fmus->mpfont_filename, 1); //fmus->edit_mode); // edit_mode = load kerning table
    if (e != FMUS_E_OK)
      return e;
    printf("\n");
  }
    
  if (fmus->query_only)
    return FMUS_E_OK;
  
  e = apply_pre_render_transformations(fmus);
  if (e != FMUS_E_OK)
    return e;
  
  if (!fmus->edit_mode) {
    
    fmus->text = FMUS_MALLOCMSG(sizeof(fmus_text_t), "fmus->text");
    fmus_text_init(fmus, fmus->text);
    
    e = fmus_main_do_txtr(fmus, &(fmus->mpfont));
    
    if (!(fmus->skip_kerning)) {
      if (fmus->mpfont.kerning_entries) {
        if (fmus->kerning)
          FMUS_FREEMSG(fmus->kerning, "fmus->kerning");
        fmus->kerning = NULL;
        fmus->mpfont.kerning_entries = 0;
      }
      //e = get_kerning_from_freetype(fmus, fmus->text, &(fmus->mpfont), fmus->font_face_ix);
      //if (e != FMUS_E_OK)
      //  return e;
    }
    
    fmus_text_finish(fmus->text);
    FMUS_FREEMSG(fmus->text, "fmus->text");
    fmus->text = NULL;
    if (e != FMUS_E_OK)
      return e;

  }
  
  e = apply_post_render_transformations(fmus);
  if (e != FMUS_E_OK)
    return e;
  
  // export kerning in edit mode only (FIXME)
  //do_kerning = fmus->edit_mode;
  
  if (!(fmus->no_compact_kerning))
    update_kerning_indices (fmus, &(fmus->mpfont));
  
  e = export_mpfont(fmus, &(fmus->mpfont), 1, &newfont, &newfont_len);
  if (e != FMUS_E_OK)
    return e;
  
  if (fmus->output_font_filename == NULL)
    fmus->output_font_filename = "fontimus.FONT";
    
  e = fmus_write_file(fmus,
                      fmus->output_font_filename,
                      fmus->overwrite,
                      newfont,
                      newfont_len,
                      0); // no mkdir

  if (e == FMUS_E_OK)
    FMUS_MPRINTF(fmus->tw, "Wrote output .FONT to %s (%lu bytes).\n",
                 fmus->output_font_filename, (u64_t) newfont_len);

  FMUS_FREEMSG(newfont, "FONT");
  newfont = NULL;
  newfont_len = 0;
  
  if (e != FMUS_E_OK)
    return e;
    
  //e = fmus_get_kerning (fmus, &(fmus->kerning));

  return FMUS_E_OK;

}

static fmus_err_t fmus_main_do_txtr (fmus_t *fmus, fmus_mpfont_t *mpfont) {
  
  fmus_err_t e;
  s32_t font_face_ix;
  // [sheet index][layer index][width][height]:
  // layer 0 -- fill
  // layer 1 -- outline
  // layer 2 -- bonus layer for tiling (using white blocks)
  u8_t (*sheets)[3][SHEET_MAX_SIZE][SHEET_MAX_SIZE];
  s32_t max_x;
  s32_t max_y;
  u8_t *txtr;
  size_t txtr_len;
  char *texture_filename;
  char default_texture_filename_32[] = "XXXXXXXX-fontimus.TXTR";
  char default_texture_filename_64[] = "XXXXXXXXXXXXXXXX-fontimus.TXTR";
  u8_t using_default_texture_filename = 0;
  u64_t txid_64;
  u32_t txid_size;
  u32_t i;
  
  sheets = NULL;
  
  font_face_ix = fmus->font_face_ix;
    
  e = fmus_text_load_font (fmus,
                           fmus->text,
                           fmus->font_filename,
                           font_face_ix);
  
  if (e != FMUS_E_OK)
    return e;
    
  //printf("\n");
  
  //if (fmus->font_px_size < 0.0) {
  //  FMUS_EPRINTF(fmus->tw, "Error: BUG: font_px_size uninitialised in fmus_main_do_txtr()\n");
  //  return FMUS_E_BUG;
  //}
  
  if (!fmus->edit_mode && !fmus->do_not_load && !fmus->override_op_font_name)
    set_mpfont_name_from_freetype(fmus->text, font_face_ix, &(mpfont->name));
    
  fmus->num_bitmaps = mpfont->num_glyphs;
  if (0 != fmus->num_bitmaps)
    fmus->bitmaps = FMUS_MALLOCMSG (sizeof(u8_t *) * (fmus->num_bitmaps), "fmus->bitmaps");
  e = fmus_draw_bitmaps (fmus,
                         mpfont,
                         font_face_ix,
                         mpfont->font_size, //fmus->font_px_size,
                         &(fmus->bitmaps));
  if ((e != FMUS_E_OK) && (fmus->bitmaps != NULL)) {
    for (i=0; i < (mpfont->num_glyphs); i++) {
      FMUS_FREEMSG((fmus->bitmaps)[i], "glyph bitmap out");
      (fmus->bitmaps)[i] = NULL;
    }
    FMUS_FREEMSG((fmus->bitmaps), "fmus->bitmaps");
    (fmus->bitmaps) = NULL;
  }

  if (e != FMUS_E_OK)
    return e;
  
  // [4][3][SHEET_MAX_SIZE][SHEET_MAX_SIZE]
  sheets = FMUS_MALLOCMSG(4 * 3 * SHEET_MAX_SIZE * SHEET_MAX_SIZE, "sheets");
  memset(sheets, 0, 4 * 3 * SHEET_MAX_SIZE * SHEET_MAX_SIZE);

  printf("\n");
 
  e = try_sheet_sizes (fmus, sheets, &max_x);
  if (e != FMUS_E_OK) {
    FMUS_FREEMSG(sheets, "sheets");
    FMUS_EPRINTF(fmus->tw, "Error building new texture: "
                          "Glyphs will not fit on the texture at this size. "
                          "Try reducing the font pixel size and trying again.\n");
    return e;
  }
  max_y = max_x;

  //get_final_sheet_dimensions(i, sheets, &max_x, &max_y);
  //max_x = max_y = i;
  
  FMUS_MPRINTF(fmus->tw, "Using sheet size %i.\n", max_x);
  
  // we have the sheets
  // use them to create the output texture!
  e = export_texture(fmus,
                     mpfont->tx_fmt,
                     max_x,
                     max_y,
                     sheets,
                     &txtr,
                     &txtr_len);
  FMUS_FREEMSG(sheets, "sheets");
  sheets = NULL;
  if (e != FMUS_E_OK)
    return e;
  
  texture_filename = NULL;
  
  if (fmus->texture_filename != NULL) {
    texture_filename = fmus->texture_filename;
  } else {
    using_default_texture_filename = 1;
    if (fmus->mpfont.version < 5)
      texture_filename = default_texture_filename_32;
    else
      texture_filename = default_texture_filename_64;
  }
  
  txid_size = ((fmus->mpfont.version == 5) ? 8 : 4);
  txid_64 = fmus->mpfont.txid_64;
  
//printf("txid_size=%u\n", txid_size);
  
  if (using_default_texture_filename) {
    for (i=0; i < txid_size * 2; i++) {
      u8_t nybble;
      // get nybble
      if (fmus->mpfont.version < 5)
        nybble = ((txid_64 >> 28) & 0xf);
      else
        nybble = ((txid_64 >> (28 + 32)) & 0xf);
      // render nybble
      if (nybble < 10)
        texture_filename[i] = '0' + nybble;
      else
        texture_filename[i] = 'a' + (nybble-10);
      // shift
      if (fmus->mpfont.version < 5)
        txid_64 = (txid_64 << 4) & 0xfffffff0;
      else
        txid_64 = (txid_64 << 4) & 0xfffffffffffffff0;
    }
  }

  e = fmus_write_file(fmus,
                      texture_filename,
                      fmus->overwrite,
                      txtr,
                      txtr_len,
                      0); //mkpath);

  FMUS_FREEMSG(txtr, "TXTR");
  
  if (e != FMUS_E_OK)
    return e;
    
  FMUS_MPRINTF(fmus->tw, "\nWrote output .TXTR to %s (%lu bytes).\n",
               texture_filename, (u64_t) txtr_len);
  
  return FMUS_E_OK;
  
}

static fmus_err_t try_sheet_sizes(fmus_t *fmus,
                                  u8_t (*sheets)[3][SHEET_MAX_SIZE][SHEET_MAX_SIZE],
                                  s32_t *sheet_size_out) {
  
  size_t cursor_pos = 0;
  u32_t i;
  u8_t current_sheet;
  u32_t glyph_ix=0; // current glyph
  u32_t j;
  
  *sheet_size_out = 0;
  current_sheet = 0;
  
  gclip_slice_printf(fmus->tw, &cursor_pos, 0, "Trying sheet sizes: ");
  
  for (i=SHEET_MIN_SIZE; i<SHEET_MAX_SIZE; i+=8) { // FIXME: SHEET_MAX_W
    fmus_err_t e;
    if (cursor_pos >= fmus->tw) // workaround for bug
      cursor_pos=0;
    if (i > SHEET_MIN_SIZE)
      gclip_slice_printf(fmus->tw, &cursor_pos, 0, ", ");
    if (cursor_pos >= fmus->tw) // workaround for bug
      cursor_pos=0;
    gclip_slice_printf(fmus->tw, &cursor_pos, 0, "%i", i);
    if (cursor_pos >= fmus->tw) // workaround for bug
      cursor_pos=0;
    fflush(stdout);
    e = place_glyphs (fmus,
                      fmus->mpfont.tx_fmt,
                      i,
                      sheets,
                      &glyph_ix,
                      &current_sheet);
    if (e != FMUS_E_OK) {
      if (e == FMUS_E_FT_NO_GLYPH_SOLUTION) {
        //FMUS_FREEMSG(txtr, "TXTR");
        continue;  // try again with larger sheet
      } else {
        return e; // fatal, give up
      }
    } else
      break;       // success
  }
  
  printf("\n");
  
  if (i >= SHEET_MAX_SIZE) {
    return FMUS_E_FT_MAX_SHEET_SIZE;
  }
  
  *sheet_size_out = i;
  
  // divide tx coordinates through by final sheet size
  for (j=0;j<fmus->mpfont.num_glyphs;j++) {
    fmus->glyphs[j].tx_left /= ((float) i);
    fmus->glyphs[j].tx_top /= ((float) i);
    fmus->glyphs[j].tx_right /= ((float) i);
    fmus->glyphs[j].tx_bottom /= ((float) i);
  }

  return FMUS_E_OK;
}

// converts from the font size used in the mpfont
// files to a font pixel size as used by freetype
// (approximate)
#define RS_VS_FT_RATIO (32.0f / 18.0f)

static fmus_err_t render_glyph (fmus_t *fmus,
                                fmus_text_t *text,
                                s32_t face_ix,
                                u16_t utf16,
                                float px_size,
                                s32_t *width_out,
                                s32_t *height_out,
                                u8_t **bitmap_out,
                                s32_t *baseline_y_out,
                                s32_t *advance_out,
                                s32_t *left_bearing_out) {
                                
  fmus_err_t e;
                                  
  e = fmus_text_render (fmus,
                        text,
                        face_ix,
                        utf16,
                        (px_size * RS_VS_FT_RATIO), // mpfont size -> freetype pixel size
                        // out:
                        width_out,
                        height_out,
                        bitmap_out,
                        baseline_y_out,
                        advance_out,
                        left_bearing_out);
                        
  if (e == FMUS_E_FT_GET_CHAR_IX) {
    // char not present in font
    *width_out = 0;
    *height_out = 0;
    e = FMUS_E_FT_GET_CHAR_IX;
  }

  if (e != FMUS_E_OK)
    return e;
    
//printf("w %i, h %i\n", width, height);
                        
  if (*width_out & 0xfffffff80) {
    // fail
    FMUS_EPRINTF(fmus->tw,
                 "Error drawing bitmaps: FreeType rendered character is too wide "
                 "(%i pixels, max. 127).\n",
                 *width_out);
    return FMUS_E_FT_RENDERED_CHAR_TOO_WIDE;
  }
  
  if (*height_out & 0xffffff80) {
    // fail
    FMUS_EPRINTF(fmus->tw,
                 "Error drawing bitmaps: FreeType rendered character is too tall "
                 "(%i pixels, max. 127).\n",
                 *height_out);
    return FMUS_E_FT_RENDERED_CHAR_TOO_TALL;
  }
  
  if ((*baseline_y_out > 127) || (*baseline_y_out < -128)) {
    // fail
    FMUS_EPRINTF(fmus->tw,
                 "Error drawing bitmaps: FreeType rendered character has a bad "
                 "baseline_y value (%i pixels, must be -128 <= y <= 127).\n",
                 *baseline_y_out);
    return FMUS_E_FT_RENDERED_CHAR_BASELINE_Y;
  }
  
  if ((*left_bearing_out > 127) || (*left_bearing_out < -128)) {
    // fail
    FMUS_EPRINTF(fmus->tw,
                 "Error drawing bitmaps: FreeType rendered character has a bad "
                 "left_bearing value (%i pixels, must be -128 <= y <= 127).\n",
                 *left_bearing_out);
    return FMUS_E_FT_RENDERED_CHAR_L_BEARING;
  }
  return FMUS_E_OK;
}

static fmus_err_t fmus_draw_bitmaps (fmus_t *fmus,
                                     fmus_mpfont_t *mpfont,
                                     s32_t face_ix,
                                     float px_size,
                                     u8_t ***bitmaps) {
  
  fmus_err_t e;
  u32_t i;
  u32_t num_missing_glyphs=0;
  
  e = FMUS_E_OK;
  
  for (i=0; i < (mpfont->num_glyphs); i++) {
    
    s32_t baseline_y=0;
    s32_t width=0, height=0;
    s32_t advance=0;
    s32_t left_bearing=0;
                          
    if (fmus->glyphs[i].utf16 != 0) {
                        // in:
      e = render_glyph (fmus,
                        fmus->text,
                        face_ix,
                        fmus->glyphs[i].utf16,
                        px_size,
                        // out:
                        &width,
                        &height,
                        (*bitmaps) + i,
                        &baseline_y,
                        &advance,
                        &left_bearing);
                        
      if (e == FMUS_E_FT_GET_CHAR_IX) {
        num_missing_glyphs++;
        e = FMUS_E_OK;
      }
                        
      if (e != FMUS_E_OK)
        return e;
                        
    }
    
    // write new information into existing structure
    fmus->glyphs[i].height = 0xff & height;
    
    // so we take freetype's advance value,
    // subtract the glyph width. that will give us
    // (ft_advance - glyph_width) = lpad + rpad
    //if (left_bearing < 0)
    //  fmus->glyphs[i].lsb = 0;
    //else
    //  fmus->glyphs[i].lpad = 0xff & left_bearing;
    fmus->glyphs[i].lsb = 0xff & left_bearing;
    fmus->glyphs[i].layout_width = 0xff & width;
    fmus->glyphs[i].rsb = 0xff & (advance - (((s32_t) fmus->glyphs[i].lsb) + width));
    fmus->glyphs[i].display_width = fmus->glyphs[i].layout_width;
    fmus->glyphs[i].vert_offset = 0x7f & baseline_y;
      
  }
  
  if (num_missing_glyphs)
    FMUS_WPRINTF(fmus->tw, "Warning: %u characters did not have glyphs in this TTF/OTF file.", num_missing_glyphs);
  
  return e;
  
}

static fmus_err_t init_fmus(fmus_t *fmus) {
  memset(fmus, 0, sizeof(fmus_t));
  fmus->font_px_size = -1.0; //32.0;
  return FMUS_E_OK;
}

static void finish_fmus(fmus_t *fmus) {
  u32_t i;
  //if (NULL != fmus->text->font.file_contents)
  //  FMUS_FREEMSG(fmus->text->font.file_contents, "fmus_read_file payload");
  //fmus->text->font.file_contents = NULL;
  if (fmus->glyphmods != NULL)
    FMUS_FREEMSG(fmus->glyphmods, "glyphmods");
  fmus->glyphmods = NULL;
  fmus->glyphmods_alloced=0;
  fmus->glyphmods_fill=0;
  if (fmus->bitmaps) {
    for (i=0;i<fmus->num_bitmaps;i++) {
      if (fmus->bitmaps[i] != NULL)
        FMUS_FREEMSG(fmus->bitmaps[i], "glyph bitmap out");
      fmus->bitmaps[i] = NULL;
    }
    FMUS_FREEMSG(fmus->bitmaps, "bitmaps");
    fmus->bitmaps = NULL;
  }
  if (NULL != fmus->glyphs) {
    //for (i=0;i<mpfont->num_glyphs;i++) {
    FMUS_FREEMSG(fmus->glyphs, "fmus->glyphs");
    //  fmus->
    //}
    fmus->glyphs = NULL;
  }
  if (fmus->kerning != NULL) {
    FMUS_FREEMSG(fmus->kerning, "fmus->kerning");
    fmus->kerning = NULL;
  }
  if (NULL != fmus->mpfont.name) {
    FMUS_FREEMSG(fmus->mpfont.name, "mpfont->name");
    fmus->mpfont.name = NULL;
  }
}

static s32_t fmus_cli_get_term_width (void) {
  /* get terminal width if the system supports ioctls */
#ifndef GOLF_NOIOCTL
#ifdef TIOCGWINSZ
  struct winsize wsize;
  if (!(ioctl (0, TIOCGWINSZ, &wsize)))
    return wsize.ws_col;
#endif
#endif
  return -1;
}

#ifdef FMUS_DBGMEM
s32_t fmus_dbgmem_get_allocs() { return mallocs; }
s32_t fmus_dbgmem_get_frees()  { return frees;   }
#endif

#ifdef FMUS_DBGMEM2
void *fmus_realloc (void *p, size_t size, u8_t zero, const char *label) {
#else
void *fmus_realloc (void *p, size_t size, u8_t zero) {
#endif // all just to shut the compiler up
  void *q=NULL;
  if (!size) {
    printf("Error: realloc(0)\n");
    exit(FMUS_E_MALLOC_0);
  }
  q=realloc(p, size);
//if (!p) {
//printf("realloc(%p, %" FMUS_FMT_SIZET ")\n", p, size);
//}
#ifdef FMUS_DBGMEM
  if (!p)
    mallocs++;
#endif
#ifdef FMUS_DBGMEM2
  printf("%p _A_ [%s] (%p, %" FMUS_FMT_SIZET ", zero=%u)\n",
               q, label, p, size, zero);
#endif
  if (!q) {
    printf("realloc(%p, %" FMUS_FMT_SIZET ") failed.\n", p, size);
    exit(FMUS_E_MALLOC);
  }
  if (zero)
    memset(q, 0, size);
  return q;
}

#ifdef FMUS_DBGMEM2
void fmus_free (void *p, const char *label) {
#else
void fmus_free (void *p) {
#endif
  if (p) {
//printf("%p freed\n", p);
#ifdef FMUS_DBGMEM 
    frees++;
#endif
#ifdef FMUS_DBGMEM2
    printf("%p _F_ [%s]\n", p, label);
#endif
    free(p);
  }
}


static s32_t find_max_occupied_x_coord (u8_t curr_sheet_L2[SHEET_MAX_SIZE][SHEET_MAX_SIZE],
                                        s32_t test_sheet_size) {
  s32_t x, y;
  for (x=test_sheet_size-1; x >= 0; x--) {
    for (y=test_sheet_size-1; y >= 0; y--) {
      if (curr_sheet_L2[x][y] == 0xff)
        return x;
    }
  }
  return 0;
}

static s32_t find_max_occupied_y_coord (u8_t curr_sheet_L2[SHEET_MAX_SIZE][SHEET_MAX_SIZE],
                                        s32_t test_sheet_size) {
  s32_t x, y;
  for (y=test_sheet_size-1; y >= 0; y--) {
    for (x=test_sheet_size-1; x >= 0; x--) {
      if (curr_sheet_L2[x][y] == 0xff)
        return y;
    }
  }
  return 0;                  
}

static u8_t test_pixel_block (s32_t glyph_w,
                              s32_t glyph_h,
                              s32_t test_x,
                              s32_t test_y,
                              s32_t test_sheet_size,
                              u8_t curr_sheet_L2[SHEET_MAX_SIZE][SHEET_MAX_SIZE]) {
  s32_t x, y;
//printf("glyph_w = %i, glyph_h = %i\n", glyph_w, glyph_h);
  for (x=0; x<glyph_w; x++) {
    for (y=0; y<glyph_h; y++) {
      if (x + test_x >= test_sheet_size) {
        // gone off edge
        return 1;
      }
      if (y + test_y >= test_sheet_size) {
        // gone off edge
        return 1;
      }
      if (0xff == curr_sheet_L2[test_x + x][test_y + y]) { // L2 = tiling layer
        // collided
        return 1;
      }
    }
  }
  
  return 0;
  
}

static fmus_err_t place_glyphs (fmus_t *fmus,
                                u32_t tx_fmt,
                                s32_t test_sheet_size,
                                u8_t (*sheets)[3][SHEET_MAX_SIZE][SHEET_MAX_SIZE],
                                u32_t *glyph_ix_inout,
                                u8_t *current_sheet_inout) {
  
  u32_t i;
  //u8_t outline;
  fmus_err_t e;
  // [sheet index][layer index][width][height]
  // layer 0 -- fill
  // layer 1 -- outline
  // layer 2 -- bonus layer for tiling (using white blocks)
  u8_t num_sheets;
  u8_t current_sheet;
  s32_t test_x;
  s32_t test_y;
  //float test_sheet_size_f;
  float test_x_f;
  float test_y_f;
  u8_t oline=0;
  u32_t glyph_ix_orig;
  s32_t max_x, max_y;
  
  e = FMUS_E_OK;
  
  if (test_sheet_size > SHEET_MAX_SIZE) {
    return FMUS_E_FT_MAX_SHEET_SIZE;
  }
  
//printf("[1]\n");
  
  if ((tx_fmt == 0) || (tx_fmt == 1))
    num_sheets = 1;
  else if (tx_fmt == 2)
    num_sheets = 4;
  else
    num_sheets = 2;
  
  if ((tx_fmt == 1) || (tx_fmt == 3) || (tx_fmt == 8))
    oline = 1; // extra space around glyphs if we're doing outlines
  
  // resume from where we left off
  glyph_ix_orig = *glyph_ix_inout;
  //current_sheet = 0; // set to inout later
  
//printf("glyph_ix_inout = %u\n", *glyph_ix_inout);

  if (NULL == fmus->bitmaps) {
    FMUS_EPRINTF(fmus->tw, "\nError: BUG: place_glyphs(): fmus->bitmaps is NULL.\n");
    return FMUS_E_BUG;
  }
  
  // continue where we (maybe) left off from previous call if the previous sheet was too small
  for (i = glyph_ix_orig;
       i < fmus->mpfont.num_glyphs;
       i++, (*glyph_ix_inout)++) {
    
    u8_t collided;
    s32_t glyph_w;
    s32_t glyph_h;
    s32_t x, y;

    glyph_w = (fmus->glyphs[i].layout_width);
    glyph_h = fmus->glyphs[i].height;

    if ((glyph_h <= 0) || (glyph_w <= 0) || (NULL == fmus->bitmaps[i])) {
      // blank glyph!
      //(*glyph_ix_inout)++; // set this to point to next glyph so we start in the right place next call
      continue;
    }
  
    current_sheet = *current_sheet_inout;
    
    // scan tiling layer to discover current progress
    max_x = find_max_occupied_x_coord(sheets[current_sheet][2], test_sheet_size);
    max_y = find_max_occupied_y_coord(sheets[current_sheet][2], test_sheet_size);
    
//printf("progress: max_x = %i, max_y = %i\n", max_x, max_y);
    
    // aim to keep the generated bitmap square,
    // so if it's wider than it is tall, scan
    // downwards first; if it's taller than it is
    // wide, scan rightwards first
    collided = 1;
    if (max_x > max_y) {
      // scan downwards
      for (test_x = 0; test_x < test_sheet_size; test_x++) {
        for (test_y = 0; test_y < test_sheet_size; test_y++) {
          // test block of pixels having origin (test_x, test_y)
          collided = test_pixel_block (glyph_w + 2 + (oline ? 2 : 0),
                                       glyph_h + 2 + (oline ? 2 : 0),
                                       test_x,
                                       test_y,
                                       test_sheet_size,
                                       sheets[current_sheet][2]); // tiling layer
          if (!collided)
            break;
        }
        if (!collided)
          break;
      }
    } else {
      // scan rightwards
      for (test_y = 0; test_y < test_sheet_size; test_y++) {
        for (test_x = 0; test_x < test_sheet_size; test_x++) {
//printf("test_x=%i, test_y=%i\n", test_x, test_y);
          // test block of pixels having origin (test_x, test_y)
          collided = test_pixel_block (glyph_w + 2 + (oline ? 2 : 0),
                                       glyph_h + 2 + (oline ? 2 : 0),
                                       test_x,
                                       test_y,
                                       test_sheet_size,
                                       sheets[current_sheet][2]); // tiling layer
          if (!collided)
            break;
        }
        if (!collided)
          break;
      }
    }
    
    if (collided) {
      // eep, no solution
      // do not update glyph_ix_inout or current_sheet_inout, since we
      // will want to try them both again next call ...
      return FMUS_E_FT_NO_GLYPH_SOLUTION;
    }
    
    //*(glyph_ix_inout)++;
    
    // OK. have a clear point; (test_x, test_y).
    // we need to copy to the fill layer,
    // potentially to the outline layer,
    // and also to the "bonus" tiling layer.
    
    // calculate the left/right/top/bottom float texture coordinates
    //test_sheet_size_f = (float) test_sheet_size;
    test_x_f = (float) (test_x + 1);
    test_y_f = (float) (test_y + 1);
    //if (!oline) {
    // we don't know how big the final sheet size will be,
    // so we'll temporarily have to keep the integer sizes around
    // and do a pass at the end to divide them by the correct size
    /*
    fmus->glyphs[i].tx_left = test_x_f / test_sheet_size_f;
    fmus->glyphs[i].tx_top = test_y_f / test_sheet_size_f;
    fmus->glyphs[i].tx_right = (test_x_f + ((float) (glyph_w + (oline ? 2 : 0)))) / test_sheet_size_f;
    fmus->glyphs[i].tx_bottom = (test_y_f + ((float) (glyph_h + (oline ? 2 : 0)))) / test_sheet_size_f;
    */
    fmus->glyphs[i].tx_left = test_x_f;
    fmus->glyphs[i].tx_top = test_y_f;
    fmus->glyphs[i].tx_right = (test_x_f + ((float) (glyph_w + (oline ? 2 : 0))));
    fmus->glyphs[i].tx_bottom = (test_y_f + ((float) (glyph_h + (oline ? 2 : 0))));
    //}
//printf("G %u (%i x %i) at (%i, %i) (L %f, T %f, R %f, B %f)\n",
//       i, glyph_w, glyph_h, test_x, test_y,
//       fmus->glyphs[i].tx_left,
//       fmus->glyphs[i].tx_top,
//       fmus->glyphs[i].tx_right,
 //      fmus->glyphs[i].tx_bottom);
    
    if (!oline) { // no outline
      for (y=0;y<glyph_h;y++) {
        s32_t px_y = test_y + y + 1; // +1 takes into account 1 px spacing
        for (x=0;x<glyph_w;x++) {
          s32_t px_x = test_x + x + 1; // +1 again
          sheets[current_sheet][2][px_x][px_y] = 0xff; // fill on tiling layer
          // copy bitmap to fill layer
          sheets[current_sheet][0][px_x][px_y] = fmus->bitmaps[i][x + (y * glyph_w)];
        }
      }
    } else { // outline mode
      for (y=0;y<(glyph_h + 2);y++) {
        s32_t px_y = test_y + y + 1; // +1 takes into account 1 px spacing
        for (x=0;x<(glyph_w + 2);x++) {
          s32_t px_x = test_x + x + 1; // +1 again
          sheets[current_sheet][2][px_x][px_y] = 0xff; // fill on tiling layer
          // only copy to fill layer if we aren't in
          // the outline gap ... it's fiddly:
          if ((x>0) && (y>0) && (x <= glyph_w) && (y <= glyph_h))
            sheets[current_sheet][0][px_x][px_y] = fmus->bitmaps[i][(x-1) + ((y-1) * glyph_w)];
          // if we're black and x+1 is white, current=white
          // if we're black and x-1 is white, current=white
          // if we're black and y+1 is white, current=white
          // if we're black and y-1 is white, current=white
          // only interested in plotting a white pixel when the current
          // pixel is black (or out of bounds of the actual
          // rendered glyph bitmap in fmus->bitmaps which is 2px slimmer/shorter),
          // since the outline is always plotted onto
          // the black pixels around the edge of the white filled pixels
          if ((x==0)
              || (y==0)
              || (x==(glyph_w+1))
              || (y==(glyph_h+1))
              || (fmus->bitmaps[i][(x-1) + ((y-1) * glyph_w)] == 0)) { // so, only if black
            if ((x>1) && (y>0) && (y<(glyph_h+1))) {
              // check x-1
              if (255 == fmus->bitmaps[i][(x-2) + ((y-1) * glyph_w)])
                sheets[current_sheet][1][px_x][px_y] = 255;
            }
            if ((x < (glyph_w)) && (y>0) && (y<(glyph_h+1))) {
              // check x+1
              if (255 == fmus->bitmaps[i][x + ((y-1) * glyph_w)])
                sheets[current_sheet][1][px_x][px_y] = 255;
            }
            if ((y>1) && (x>0) && (x<(glyph_w+1))) {
              // check y-1
              if (255 == fmus->bitmaps[i][(x-1) + ((y-2) * glyph_w)])
                sheets[current_sheet][1][px_x][px_y] = 255;
            }
            if ((y < (glyph_h)) && (x>0) && (x<glyph_w+1)) {
              // check y+1
              if (255 == fmus->bitmaps[i][(x-1) + (y * glyph_w)])
                sheets[current_sheet][1][px_x][px_y] = 255;
            }
          }
        }
      }
    }
    
    fmus->glyphs[i].rgba_chnl_ix = current_sheet;
 
    // distribute glyphs across different sheets
    current_sheet++;
    if (current_sheet == num_sheets)
      current_sheet=0;
      
    *current_sheet_inout = current_sheet; // preserve this for next call
  }
   
  return e;
  
}

static fmus_err_t export_texture(fmus_t *fmus,
                                 u32_t tx_fmt,
                                 s32_t sheet_w,
                                 s32_t sheet_h,
                                 u8_t sheets[4][3][SHEET_MAX_SIZE][SHEET_MAX_SIZE],
                                 u8_t **txtr_out,
                                 size_t *txtr_len_out) {
                                   
  u8_t header[20];
  u8_t palette[32];
  u8_t *blockmap;
  s32_t b;
  u32_t dst_ix;
  
  // FIXME: mipmaps = 1
  memcpy(header, "\x00\x00\x00\x04wwhh\x00\x00\x00\x01pppp\x00\x01\x00\x10", 20);
  
  if (sheet_w & 0xffff0007) {
    FMUS_EPRINTF(fmus->tw, "Error exporting texture: BUG: Bad sheet width (%i).\n", sheet_w);
    return FMUS_E_BUG;
  }
  if (sheet_h & 0xffff0007) {
    FMUS_EPRINTF(fmus->tw, "Error exporting texture: BUG: Bad sheet height (%i).\n", sheet_h);
    return FMUS_E_BUG;
  }
  /*
  if (tx_fmt == 4) {
    FMUS_EPRINTF(fmus->tw, "Error exporting texture: BUG: Texture format 0x4 (with copies) "
                           "is not currently "
                           "supported.\n");
    return FMUS_E_MPF_TX_FMT;
  }
  */
  if (tx_fmt > 4) {
    FMUS_EPRINTF(fmus->tw, "Error exporting texture: BUG: Bad texture format (%i).\n", tx_fmt);
    return FMUS_E_BUG;
  }
  
  write_u16_bigendian(header+4, 0xffff & sheet_w);
  write_u16_bigendian(header+6, 0xffff & sheet_h);
  write_u32_bigendian(header+12, 0x2); // FIXME: palette format
  memset(palette, 0, 32); // FIXME: examine an actual font texture and see what's in this
  
  // 8x8 blocks and 4 bpp
  *txtr_len_out = 20 + 32 + ((sheet_w * sheet_h) / 2);
//printf("mallocing %u bytes\n", *txtr_len_out);
  *txtr_out = FMUS_MALLOCMSG(*txtr_len_out, "TXTR");
  
  memcpy(*txtr_out, header, 20);
  memcpy(*txtr_out + 20, palette, 32);
  
  blockmap = *txtr_out + 52;
  dst_ix=0;
  
  for (b=0; b<(sheet_w * sheet_h) / 64; b++) {
    
    s32_t block_origin_x;
    s32_t block_origin_y;
    s32_t x,y;
    
    block_origin_x = 8 * (b % (sheet_w / 8));
    block_origin_y = 8 * (b / (sheet_w / 8));
    
    for (y=0;y<8;y++) {
      for (x=0;x<8;x+=2) {
        u8_t px;
        s32_t tx, ty;
        tx = x + block_origin_x;
        ty = y + block_origin_y;
        px=0; // 2 x four bits
        switch (tx_fmt) {
          case 0:
            px |= 0x10 & (((u32_t) sheets[0][0][tx][ty]) << 4);         // layer 0 = fill
            px |= 0x1 & ((((u32_t) sheets[0][0][tx+1][ty])));
            break;
          case 1:
            px |= 0x10 & (((u32_t) sheets[0][0][tx][ty]) << 4);       // layer 0 = fill
            px |= 0x1 & (((u32_t) sheets[0][0][tx+1][ty])); 
            px |= 0x20 & (((u32_t) sheets[0][1][tx][ty]) << 4);       // layer 1 = outlines
            px |= 0x2 & (((u32_t) sheets[0][1][tx+1][ty])); 
            break;
          case 2:
            px |= 0x10 & (((u32_t) sheets[0][0][tx][ty]) << 4);       // layer 0 = fill sheet 0
            px |= 0x1 & (((u32_t) sheets[0][0][tx+1][ty])); 
            px |= 0x20 & (((u32_t) sheets[1][0][tx][ty]) << 4);       // layer 1 = fill sheet 1
            px |= 0x2 & (((u32_t) sheets[1][0][tx+1][ty]));
            px |= 0x40 & (((u32_t) sheets[2][0][tx][ty]) << 4);       // layer 2 = fill sheet 2
            px |= 0x4 & (((u32_t) sheets[2][0][tx+1][ty])); 
            px |= 0x80 & (((u32_t) sheets[3][0][tx][ty]) << 4);       // layer 3 = fill sheet 3
            px |= 0x8 & (((u32_t) sheets[3][0][tx+1][ty])); 
            break;
          case 3:
            px |= 0x10 & (((u32_t) sheets[0][0][tx][ty]) << 4);       // layer 0 = fill sheet 0
            px |= 0x1 & (((u32_t) sheets[0][0][tx+1][ty])); 
            px |= 0x20 & (((u32_t) sheets[0][1][tx][ty]) << 4);       // layer 1 = outline sheet 0
            px |= 0x2 & (((u32_t) sheets[0][1][tx+1][ty]));
            px |= 0x40 & (((u32_t) sheets[1][0][tx][ty]) << 4);       // layer 2 = fill sheet 1
            px |= 0x4 & (((u32_t) sheets[1][0][tx+1][ty])); 
            px |= 0x80 & (((u32_t) sheets[1][1][tx][ty]) << 4);       // layer 3 = outline sheet 1
            px |= 0x8 & (((u32_t) sheets[1][1][tx+1][ty])); 
            break;
          case 4:
            px |= 0x10 & (((u32_t) sheets[0][0][tx][ty]) << 4);       // layer 0 = fill sheet 0
            px |= 0x1 & (((u32_t) sheets[0][0][tx+1][ty])); 
            px |= 0x20 & (((u32_t) sheets[1][0][tx][ty]) << 4);       // layer 1 = fill sheet 1
            px |= 0x2 & (((u32_t) sheets[1][0][tx+1][ty]));
            px |= 0x40 & (((u32_t) sheets[0][0][tx][ty]) << 4);       // layer 2 = fill sheet 0
            px |= 0x4 & (((u32_t) sheets[0][0][tx+1][ty])); 
            px |= 0x80 & (((u32_t) sheets[1][0][tx][ty]) << 4);       // layer 3 = fill sheet 1
            px |= 0x8 & (((u32_t) sheets[1][0][tx+1][ty])); 
            break;
          default:
            break;
        }
//printf("dst_ix=%u\n", dst_ix);
        blockmap[dst_ix] = px;
        dst_ix++;
      } // next x
    } // next y
  } // next block
  return FMUS_E_OK;
}


static gclip_err_t gclip_slice_printf_grow(char **buf,
                                           size_t *alloced) {
  if (!buf || !(*buf) || !alloced)
    return GCLIP_E_NULL;
  (*alloced)<<=1;
  *buf = GCLIP_REALLOCMSG(*buf, *alloced, "slice printf");
  return GCLIP_E_OK;
}

gclip_err_t gclip_slice_printf(u16_t cols,          /* wrap width */
                               size_t *cursor_pos,  /* current cursor column */
                               size_t wrap_indent,  /* how much to indent new lines */
                               char *fmt,           /* fmtstg */
                               ...) {               /* etc. */
  va_list argptr;
  void *val_ptr = NULL;
  char  *val_stg = NULL;
  u32_t val_u32=0;
  s32_t val_s32=0;
#ifdef GCLIP_64
  u64_t val_u64=0;
  s64_t val_s64=0;
#endif
#ifdef GCLIP_FLT
  double val_dbl=0.0;
#endif
  size_t i;
  size_t alloced_len=64;
  u8_t mode=0;
  char *buf = NULL;
  char *buf2 = NULL;
  size_t fill=0;
  gclip_err_t e;
  u8_t broken_snprintf=0;
  char *fmtstg = NULL;
  size_t arglen=0;
  
  if (!fmt)
    return GCLIP_E_NULL;

  alloced_len = 1;
  if (NULL == (buf=GCLIP_MALLOCMSG(alloced_len, "slice printf")))
    return GCLIP_E_MALLOC;
  if (GCLIP_SCPRINTF_1ARG("%s", "test") <= 0) {
    /* hooray, a broken snprintf */
#ifdef GCLIP_VRB
    fprintf(GCLIP_ESTREAM, "slice printf: broken _scprintf/snprintf\n");
#endif
    broken_snprintf=1;
  }
  
  /* modes:
   * 0: in normal text, expecting %
   * 1: after %, expecting l or d, i, s etc.
   * 2: after %l, expecting d, i, s etc.
   */

  va_start(argptr, fmt);
  
  /* fill will point to the null terminator. */
  for (i=0; fmt[i]; i++) {
    
    /*
    void *p = NULL;
    */
    u8_t skip=0;
    fmtstg = NULL;
#ifdef GCLIP_VRB
    fprintf(GCLIP_ESTREAM, "slice printf: i=%" GCLIP_FMT_SIZET ", mode=%u\n", i, mode);
#endif
    switch (mode) {
      case 0: /* normal text, want % */
        if (fmt[i] == '%') {
          mode=1;
        } else {
          /* pass through */
          /* BEGIN CODEREP */
          while (fill+1 >= alloced_len) { /* grow if necessary */
            e = gclip_slice_printf_grow(&buf, &alloced_len);
            if (e != GCLIP_E_OK) {
              va_end(argptr);
              return e;
            }
          }
          /* END CODEREP */
          buf[fill] = fmt[i];
          fill++;
        }
        break;
      case 1: /* have %, want l or something else */
#ifdef GCLIP_64
      case 2: /* have %l, want something else */
#endif
        
        /* mode 1 or 2 -- for options with long modes: */
        switch (fmt[i]) {
          case 'l':
#ifdef GCLIP_64
            if (mode==1) {   /* % mode */
              mode=2; /* activate long mode */
              skip=1;
              break;
            }
#endif
            /* So, this code is reached either for %l without
             * GCLIP_64 or %ll with GCLIP_64. */
            /* this is an error, so pass through */
            /* BEGIN CODEREP */
            while (fill+3 >= alloced_len) { /* grow if necessary */
              e = gclip_slice_printf_grow(&buf, &alloced_len);
              if (e != GCLIP_E_OK) { /* buf already freed */
                va_end(argptr);
                return e;
              }
            }
            /* END CODEREP */
            memcpy (buf+fill, fmt+i-2, 3); /* 3 characters */
            mode = 0;
            fill += 3;
            skip = 1;
            break;
          case 'u':
          case 'i':
          case 'x':
          case 'd':
            switch (fmt[i]) {
              case 'u':
                fmtstg="%u";
#ifdef GCLIP_64
                if (mode==2)
                  fmtstg="%" GCLIP_FMT_U64;
#endif
                break;
              case 'i':
              case 'd':
                fmtstg="%d";
#ifdef GCLIP_64
                if (mode==2)
                  fmtstg="%" GCLIP_FMT_S64;
#endif
                break;
              case 'x':
                fmtstg="%x";
#ifdef GCLIP_64
                if (mode==2)
                  fmtstg="%" GCLIP_FMT_U64HEX;
#endif
                break;
              default:
                GCLIP_FREEMSG(buf, "slice printf");
                va_end(argptr);
                return GCLIP_E_BUG;
            }
            break;
          default:
            break;
        } /* switch (fmt[i]) -- modes 1 & 2 */
        
        if (mode==1) {
          if (fmt[i] == '%') { /* literal % */
            /* CODEREP */
            while (fill+1 >= alloced_len) { /* grow if necessary */
              e = gclip_slice_printf_grow(&buf, &alloced_len);
              if (e != GCLIP_E_OK) {
                va_end(argptr);
                return e;
              }
            }
            /* END CODEREP */
            buf[fill] = fmt[i];
            fill++;
            mode = 0;
          } else if (fmt[i] == 's') {
            fmtstg="%s";
          } else if (fmt[i] == 'p') {
            fmtstg = "%p";
#ifdef GCLIP_FLT
          } else if (fmt[i] == 'f') {
            fmtstg="%lf";
#endif
          }
        }

        if (skip)
          continue;
        
        if (!fmtstg) { /* unrecognised format */
          if (mode==1) {
            /* pass through 2 chars */
            arglen=2;
          } else if (mode==2) {
            /* pass through 3 chars */
            arglen=3;
          }
        } else {
          switch (fmt[i]) {
            case 'u':
            case 'x':
#ifdef GCLIP_64
              if (mode==2) { /* long mode */
                val_u64 = va_arg(argptr, u64_t);
                arglen = GCLIP_SCPRINTF_1ARG(fmtstg, val_u64);
                break;
              }
#endif
              val_u32 = va_arg(argptr, u32_t);
              arglen = GCLIP_SCPRINTF_1ARG(fmtstg, val_u32);
              break;
            case 'i':
            case 'd':
#ifdef GCLIP_64
              if (mode==2) { /* long mode */
                val_s64 = va_arg(argptr, s64_t);
                arglen = GCLIP_SCPRINTF_1ARG(fmtstg, val_s64);
                break;
              }
#endif
              val_s32 = va_arg(argptr, s32_t);
              arglen = GCLIP_SCPRINTF_1ARG(fmtstg, val_s32);
              break;
            case 's':
              val_stg = va_arg(argptr, char *);
              arglen = GCLIP_SCPRINTF_1ARG(fmtstg, val_stg);
              break;
            case 'p':
              val_ptr = va_arg(argptr, void *);
              arglen = GCLIP_SCPRINTF_1ARG(fmtstg, val_ptr);
              break;
#ifdef GCLIP_FLT
            case 'f':
              val_dbl = va_arg(argptr, double);
              arglen = GCLIP_SCPRINTF_1ARG(fmtstg, val_dbl);
              break;
#endif
          } /* end switch */
          if (broken_snprintf) {
            /* shrug, overwrite arglen with a guess */
            arglen = 512;
          }
        } /* end if valid format */
        
#ifdef GCLIP_VRB
        fprintf(GCLIP_ESTREAM,
                "slice printf: arglen[1]=%" GCLIP_FMT_SIZET "\n",
                arglen);
#endif
        
        /* alloc */
        /* CODEREP */
        while (fill+arglen+1 >= alloced_len) { /* grow if necessary */
          e = gclip_slice_printf_grow(&buf, &alloced_len);
          if (e != GCLIP_E_OK) { /* buf already freed */
            va_end(argptr);
            return e;
          }
        }
        /* END CODEREP */
        
        
        if (!fmtstg) {
          memcpy(buf+fill, fmt+i-(arglen-1), arglen);
          fill += arglen;
        } else {
          /* do the snprintf */

          switch (fmt[i]) {
            case 'u':
            case 'x':
#ifdef GCLIP_64
              if (mode==2) { /* long mode */
                GCLIP_SNPRINTF_1ARG(buf+fill, alloced_len-fill,
                                    arglen+1, fmtstg, val_u64);
                break;
              }
#endif
              GCLIP_SNPRINTF_1ARG(buf+fill, alloced_len-fill,
                                  arglen+1, fmtstg, val_u32);
              break;
            case 'i':
            case 'd':
#ifdef GCLIP_64
              if (mode==2) { /* long mode */
                GCLIP_SNPRINTF_1ARG(buf+fill, alloced_len-fill,
                                    arglen+1, fmtstg, val_s64);
                break;
              }
#endif
              GCLIP_SNPRINTF_1ARG(buf+fill, alloced_len-fill,
                                  arglen+1, fmtstg, val_s32);
              break;
            case 's':
              GCLIP_SNPRINTF_1ARG(buf+fill, alloced_len-fill,
                                  arglen+1, fmtstg, val_stg);
              break;
            case 'p':
              GCLIP_SNPRINTF_1ARG(buf+fill, alloced_len-fill,
                                  arglen+1, fmtstg, val_ptr);
              break;
#ifdef GCLIP_FLT
            case 'f':
              GCLIP_SNPRINTF_1ARG(buf+fill, alloced_len-fill,
                                  arglen+1, fmtstg, val_dbl);
              break;
#endif
          }

#ifdef GCLIP_VRB
          fprintf(GCLIP_ESTREAM,
                  "slice printf: arglen[2]=%" GCLIP_FMT_SIZET "\n",
                  arglen);
#endif
          
          if (!broken_snprintf) {
            fill += arglen;
          } else {
            fill = strlen((char *) buf); /* haha */
          }
        }
#ifdef GCLIP_VRB
        fprintf(GCLIP_ESTREAM, "slice printf: fill=%" GCLIP_FMT_SIZET "\n", fill);
#endif
        
        mode=0;
        
        break;
        
      default:
        /* FIXME: bug */
        break;
        
    } /* end switch(mode) */
    
  } /* end for i loop */
  
  /* deal with leftovers */
  switch (mode) {
    case 0:
      arglen=0;
      break; /* should be no leftovers */
    case 1:
      arglen=1; /* "%" */
      break;
    case 2:
      arglen=2; /* "%l" */
      break;
    default:
      /* FIXME: bug */
      break;
  }
  
  /* CODEREP */
  while (fill+arglen+1 >= alloced_len) { /* grow if necessary */
    e = gclip_slice_printf_grow(&buf, &alloced_len);
    if (e != GCLIP_E_OK) { /* buf already freed */
      va_end(argptr);
      return e;
    }
  }
  /* END CODEREP */
  
  if (arglen) {
    buf[fill] = '%';
    fill++;
    if (arglen==2) {
      buf[fill] = 'l';
      fill++;
    }
  }
  
  /* null-terminate output string */
  buf[fill] = 0;
  /*
  fill++;
  */
  
  va_end(argptr);
  
  e=gclip_lineslice(cols, buf, &buf2, cursor_pos, wrap_indent);
  
  GCLIP_FREEMSG(buf, "slice printf");
  if (GCLIP_E_OK != e)
    return e;
  
  fprintf(GCLIP_ESTREAM, "%s", buf2);
  gclip_free_lineslice(buf2);
  
  return GCLIP_E_OK;
  
} /* end function */

static gclip_err_t gclip_free_lineslice(char *buf) {
  if (!buf)
    return GCLIP_E_NULL;
  GCLIP_FREEMSG(buf, "lineslice buffer");
  return GCLIP_E_OK;
}

/*
#define GCLIP_LINESLICE_DEBUG
*/
// buggy
static gclip_err_t gclip_lineslice(u16_t cols,
                                   char *in,
                                   char **out,
                                   size_t *cursor_pos, /* both input and output */
                                   size_t wrap_indent) {
  u16_t chunklen;
  u8_t pass;
  size_t in_len=0;
  size_t in_off;
  size_t out_off=0;
  size_t dummy_cursor_pos=0;
  size_t cursor_pos_orig=0;
  s32_t i;
  if (!in) {
    fprintf(GCLIP_ESTREAM, "gclip_lineslice: in==NULL\n");
    return GCLIP_E_NULL;
  }
  if (!out) {
    fprintf(GCLIP_ESTREAM, "gclip_lineslice: out==NULL\n");
    return GCLIP_E_NULL;
  }
  if (!cols) {
    fprintf(GCLIP_ESTREAM, "gclip_lineslice: cols=0\n");
    return GCLIP_E_ZEROWIDTH;
  }
  if (wrap_indent >= cols) {
    fprintf(GCLIP_ESTREAM,
            "gclip_lineslice: wrap_indent(%" GCLIP_FMT_SIZET ") >= cols (%u)\n",
            wrap_indent, cols);
    return GCLIP_E_BADINDENT;
  }
  
  if (cursor_pos && (*cursor_pos >= cols)) {
    fprintf(GCLIP_ESTREAM,
            "gclip_lineslice: *cursor_pos(%" GCLIP_FMT_SIZET ") >= cols (%u)\n",
            *cursor_pos, cols);
    return GCLIP_E_BADINDENT;
  }
  in_len = strlen(in);
/*printf("%s\n", in);*/
  if (!cursor_pos)
    cursor_pos = &dummy_cursor_pos;
  cursor_pos_orig = *cursor_pos;
  /* two passes */
  for (pass=0;pass<2;pass++) {
    u8_t leftovers=0;
    in_off = 0;
    out_off = 0;
    *cursor_pos = cursor_pos_orig;
/*printf("[pass=%u]", pass);*/
    while (in_off < in_len) {
      /* apart from the very first time,
       * there has always been a newline just
       * copied to the output if the code is
       * at this point. */
      u8_t found_newline=0;
#ifdef GCLIP_LINESLICE_DEBUG
fprintf(GCLIP_ESTREAM, "[1 cp=%u]", *cursor_pos);
#endif

      if (in_off && wrap_indent) {
        /* copy the wrap_indent padding
         * to the output */
        if (pass)
          memset((*out)+out_off, ' ', wrap_indent);
        /* update output position */
        out_off += wrap_indent;
        /* update cursor position */
        *cursor_pos += wrap_indent;
#ifdef GCLIP_LINESLICE_DEBUG
fprintf(GCLIP_ESTREAM, "[2 cp=%u]", *cursor_pos);
#endif
      }
#ifdef GCLIP_LINESLICE_DEBUG
fprintf(GCLIP_ESTREAM, "[10 cp=%u]", *cursor_pos);
#endif
      /* consider a chunk: a line's worth of input */
      chunklen = cols - (cursor_pos?(*cursor_pos):0); /* the most we can display on one line */
#ifdef GCLIP_LINESLICE_DEBUG
fprintf(GCLIP_ESTREAM, "[11 cl=%u]", chunklen);
#endif
      if (in_off + chunklen >= in_len) {
        leftovers = 1; /* the final chunk, maybe */
        chunklen -= (in_off + chunklen) - in_len; /* truncate final chunk */
      }
#ifdef GCLIP_LINESLICE_DEBUG
fprintf(GCLIP_ESTREAM, "[9 cl=%u]", chunklen);
#endif
      /* scan chunk for newlines */
      for (i=0;i<chunklen;i++) {
        if (in[in_off+i] == '\n') {
          if (pass) {
            /* copy input to output */
//gclip_hexdump_2(out+out_off, i+1, 0, 16, stdout);
            memcpy(*out+out_off, in+in_off, i+1);
          }
#ifdef GCLIP_LINESLICE_DEBUG
fprintf(GCLIP_ESTREAM, "[3]");
#endif
          in_off += i+1; /* +1 to skip the newline */
          out_off += i+1; /* +1 to copy newline */
          found_newline=1;
          *cursor_pos=0;
          break;
        }
      }
      if (found_newline)
        continue;
      /* no newline was found in the chunk,
       * so search backwards for a space (unless the chunk will fit
       * on a line) */
      if (!leftovers && chunklen) {
        for (i=chunklen-1;i>=0;i--) {
          if (in[in_off+i] == ' ') {
            /* OK, we'll break the line here */
            if (pass) {
              /* copy input to output */
              memcpy (*out+out_off, in+in_off, i);
              (*out)[out_off+i] = '\n';
            }
#ifdef GCLIP_LINESLICE_DEBUG
fprintf(GCLIP_ESTREAM, "[4]");
#endif
            in_off  += i+1; /* +1 to skip the space */
            out_off += i+1; /* +1 for newline */
            found_newline=1;
            *cursor_pos=0;
            break;
          }
        }
      } else {
        /* it's leftovers, and no newline was found in them,
         * so just copy them straight to the output */
#ifdef GCLIP_LINESLICE_DEBUG
fprintf(GCLIP_ESTREAM, "[5]");
#endif
        if (pass)
          memcpy(*out+out_off, in+in_off, chunklen);
#ifdef GCLIP_LINESLICE_DEBUG
fprintf(GCLIP_ESTREAM, "[6 in_off=%u, cl=%u]", in_off, chunklen);
#endif
        /* removed: values are never read (thanks, clang) */
        /* in_off += chunklen; */
        /* found_newline=1; */
        out_off += chunklen;
        *cursor_pos += chunklen;
#ifdef GCLIP_LINESLICE_DEBUG
fprintf(GCLIP_ESTREAM, "[7 cp=%u]", *cursor_pos);
#endif
        break;
      }
      if (found_newline)
        continue;
      /* no spaces were found in the chunk (or chunk is leftovers),
       * so we have two choices:
       *   i)  just break the line here mid-word, or
       *   ii) search forwards for the next space
       *       and break the line there, exceeding
       *       the line length limit.
       * option (ii) is more complicated so we'll
       * go with (i) for now */
      /* just copy the whole chunk to the output
       * and then insert a newline */
      if (pass) {
        memcpy(*out+out_off, in+in_off, chunklen);
        (*out)[out_off+chunklen] = '\n';
      }
#ifdef GCLIP_LINESLICE_DEBUG
fprintf(GCLIP_ESTREAM, ("[8]");
#endif
      *cursor_pos = 0;
      in_off += chunklen;
      out_off += chunklen+1; /* +1 for newline */
    }
    if (!pass) {
      *out = GCLIP_MALLOCMSG(out_off+1, "lineslice buffer");
    } else {
      (*out)[out_off] = 0; /* terminator */
    }
  }
  return GCLIP_E_OK;
}
