#include "fmus_io.h"
#include "fmus.h"

static fmus_err_t fmus_mkdir (fmus_t *fmus, char *path);
static fmus_err_t fmus_mkdir_2 (fmus_t *fmus, char *path);

#ifdef FMUS_WINDOZE
//#include "Windows.h"
#include "Shlobj.h"
#define FMUS_HOMEDIR "Saved Games" FMUS_PATHSEP_S "Pad Down"
#define FMUS_ENVVAR_HOME "USERPROFILE"
#else
#define FMUS_ENVVAR_HOME "HOME"
#define FMUS_HOMEDIR ".pad-down"
#endif

FILE *fmus_fopen (u32_t terminal_width, char *filename, char *mode) {
  
  FILE *f = NULL;
  u8_t failed=0;
  char *errmsg = NULL;
  
  if (!filename) {
    printf("BUG: fmus_fopen(filename==NULL)\n");
    return NULL;
  }
  if (!mode) {
    printf("BUG: fmus_fopen(mode==NULL)\n");
    return NULL;
  }

//printf("fopen(%s)\n", filename);

#ifdef FMUS_WINDOZE
  if (fopen_s (&f, filename, mode))
    failed = 1;
#else
  if (NULL == (f = fopen (filename, mode)))
    failed = 1;
#endif
  if (failed) {
//#ifdef FMUS_WINDOZE
//    errmsg = "FIXME: no fopen error messages on windoze";
//#else
    errmsg = FMUS_STRERROR();
//#endif
    FMUS_EPRINTF(terminal_width, //7,
                 "Error: Could not open file \"%s\" (%s).\n",
                 filename, errmsg);
  }
  return f;
}



fmus_err_t fmus_read_file (fmus_t *fmus, char *filename,
                           u8_t **payload_out, size_t *len_out) {
  // *payload_out must be freed by the caller if the result is FMUS_E_OK
  
  fmus_stat_t sb;
  
  fmus_err_t ecode = FMUS_E_OK;
  FILE *fp = NULL;
  u64_t mask = 0;
  
  *payload_out = NULL;
  *len_out = 0;
  
  if (!filename) {
    FMUS_BPRINTF(fmus->tw, "BUG: fmus_read_file(filename==NULL)\n");
    return FMUS_E_NULL;
  }
  if (!payload_out) {
    FMUS_BPRINTF(fmus->tw, "BUG: fmus_read_file(payload_out==NULL)\n");
    return FMUS_E_NULL;
  }
  if (!len_out) {
    FMUS_BPRINTF(fmus->tw, "BUG: fmus_read_file(len_out==NULL)\n");
    return FMUS_E_NULL;
  }

  // stat the file once before opening it,
  // to check it isn't a directory or something
  memset(&sb, 0, sizeof(fmus_stat_t));
  if (FMUS_E_OK != (ecode = fmus_stat (fmus, filename, &sb))) {
    FMUS_EPRINTF(fmus->tw, "Error: File to be loaded is missing: \"%s\".\n", filename);
    return ecode;
  }

  if (sb.is_dir) {
    FMUS_EPRINTF(fmus->tw,
                 "Error: Could not load file \"%s\", because it is "
                 "actually a directory.\n", filename);
    return FMUS_E_FILE_NOT_FILE;
  }
  
  // then, open it
  if (NULL == (fp = fmus_fopen (fmus->tw, filename, "rb")))
    return FMUS_E_FOPEN_READ_FILE;

  // and now we stat it again
  memset(&sb, 0, sizeof(fmus_stat_t));
  if (FMUS_E_OK != (ecode = fmus_stat (fmus, filename, &sb))) {
    fclose(fp);
    return ecode;
  }

  if (sb.is_dir) {
    FMUS_EPRINTF(fmus->tw,
                 "Error: Could not load file \"%s\", because it is "
                 "actually a directory.\n", filename);
    fclose(fp);
    return FMUS_E_FILE_NOT_FILE;
  }
  
  if (sb.fsize==0) {
    fclose(fp);
    return FMUS_E_OK;
  }
  
  *payload_out = (u8_t *) FMUS_MALLOCMSG(sb.fsize, "fmus_read_file payload");
  
  if (sizeof(off_t) == 4 || sizeof(size_t) == 4)
    mask = 0x7fffffff;
  else
    mask = 0x7fffffffffffffff;

  ecode = fread_fully (fmus->tw, *payload_out, sb.fsize & mask, fp);
  fclose(fp);
  
  if (FMUS_E_OK != ecode) {
    FMUS_FREEMSG(*payload_out, "fmus_read_file payload");
    *payload_out = NULL;
    return ecode;
  }
  
  *len_out = sb.fsize;

  return ecode;
}



fmus_err_t fread_fully(//fmus_t *fmus,
                       u32_t terminal_width,
                       u8_t *buf,
                       size_t size,
                       FILE *f) {
  size_t size_read=0;
  char *errmsg = NULL;
  while (size_read<size) {
    size_t j;
    size_read += (j = fread (buf+size_read, 1, size-size_read, f));
    if (ferror(f)) {
      errmsg = FMUS_STRERROR();
      FMUS_EPRINTF(terminal_width,
                   "Error: Problem reading file: %s.\n",
                   errmsg);
      return FMUS_E_FREAD;
    } else if (feof(f)) {
      return FMUS_E_FREAD_EOF;
    }
    if (!j) {
#ifndef FMUS_WINDOZE
      usleep(1000);
#else
      Sleep(1); /* 1 ms */
#endif
    }
  }
  return FMUS_E_OK;
}

// FIXME: add follow_link parameter, choose between stat() and lstat()
fmus_err_t fmus_stat (fmus_t *fmus, char *filename, fmus_stat_t *stat_out) {
#ifdef FMUS_WINDOZE
  WIN32_FILE_ATTRIBUTE_DATA fad;
  u64_t tmpu64;
#else
  struct stat sb;
#endif
  if (!filename) {
    FMUS_BPRINTF(fmus->tw, "BUG: fmus_stat(filename==NULL)\n");
    return FMUS_E_NULL;
  }
  if (!stat_out) {
    FMUS_BPRINTF(fmus->tw, "BUG: fmus_stat(stat_out==NULL)\n");
    return FMUS_E_NULL;
  }

#ifdef FMUS_WINDOZE
  memset(&fad, 0, sizeof(WIN32_FILE_ATTRIBUTE_DATA));
#else
  memset(&sb, 0, sizeof(struct stat));
#endif

  memset(stat_out, 0, sizeof(fmus_stat_t));

#ifdef FMUS_WINDOZE
  if (!GetFileAttributesEx(filename, GetFileExInfoStandard, &fad))
    return FMUS_E_STAT;
//printf("file attributes(%s) = 0x%x\n", filename, fad.dwFileAttributes);
  if (fad.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
//printf("is_dir=1\n");
    stat_out->is_dir = 1;
  }
  tmpu64 = fad.nFileSizeHigh;
  tmpu64 <<= 32;
  tmpu64 &= 0xffffffff00000000UL;
  tmpu64 |= 0xffffffff & fad.nFileSizeLow;
  stat_out->fsize = (off_t) 0x7fffffffffffffff & tmpu64;
#else // ndef FMUS_WINDOZE
  if (lstat(filename, &sb))
    return FMUS_E_STAT;
  if ((sb.st_mode & S_IFMT) == S_IFLNK)
    stat_out->is_link = 1;
  if (stat(filename, &sb))
    return FMUS_E_STAT;
  if ((sb.st_mode & S_IFMT) == S_IFDIR)
    stat_out->is_dir = 1;

  stat_out->fsize = sb.st_size;
#endif
  return FMUS_E_OK;
}



fmus_err_t fmus_write_file (fmus_t *fmus, char *filename, u8_t overwrite,
                            u8_t *payload, size_t len, u8_t mkpath) {

  fmus_stat_t sb;

  FILE *fp = NULL;
  fmus_err_t ecode = FMUS_E_OK;
  char *symlink_text = " ";

  // general file writing scheme
  // does the file exist? if so, do we have permission to overwrite it?
  memset(&sb, 0, sizeof(fmus_stat_t));
  if (FMUS_E_OK == fmus_stat(fmus, filename, &sb)) {
    // stat succeeded: file exists
    if (overwrite) {
      // but we are cleared to overwrite it
      if (sb.is_dir || sb.is_link) {
        // but we can't, because it's a directory or symlink
#ifndef FMUS_WINDOZE
        symlink_text = " a symbolic link or ";
#endif
        FMUS_EPRINTF(fmus->tw,
                    "Error: Could not overwrite existing file \"%s\" because it is "
                    "not actually a file. It is"
                    "%s"
                    "a directory.\n",
                    filename, symlink_text);
        return FMUS_E_FILE_NOT_FILE;
      }
    } else {
      FMUS_EPRINTF(fmus->tw,
                  "Error: Refusing to overwrite file \"%s\". Use -w to override this behaviour.\n",
                  filename);
      return FMUS_E_FILE_EXISTS;
    }
  }

  if (mkpath) {
    if (FMUS_E_OK != (ecode = mkdir_for_file (fmus, filename))) {
      return ecode;
    }
  }
  
  // write the payload
  if (NULL == (fp = fmus_fopen (fmus->tw, filename, "wb"))) {
    /*
    errmsg = FMUS_STRERROR(fmus);
    FMUS_EPRINTF(fmus->tw, 7,
                 "Error: Could not write file \"%s\" (%s).\n",
                 filename, errmsg);
    */
    return FMUS_E_FOPEN_OPFILE;
  }
  ecode = fwrite_fully (fmus, payload, len, filename, fp);
  fclose(fp);
  return ecode;
}




fmus_err_t fwrite_fully(fmus_t *fmus,
                        u8_t *buf,
                        size_t size,
                        char *filename,
                        FILE *f) {
  size_t size_written=0;
  char *errmsg = NULL;
  while (size_written<size) {
    size_t j;
//printf("buf=%p\n", buf);
//printf("fwrite(%p, 1, %zu, [%s])\n", buf+size_written, size-size_written, filename);
    size_written+=(j=fwrite(buf+size_written,1,size-size_written,f));
    
    if (ferror(f)) {
      errmsg = FMUS_STRERROR();
      FMUS_EPRINTF(fmus->tw,
                   "Error: Something went wrong writing file %s (%s).\n",
                   filename, errmsg);
      return FMUS_E_FWRITE;
    } else if (feof(f)) {
      errmsg = FMUS_STRERROR();
      FMUS_EPRINTF(fmus->tw,
                   "Error: An unexpected EOF was encountered while writing file %s (%s).\n",
                   filename, errmsg);
      return FMUS_E_FREAD_EOF;
    }
    if (!j) {
#ifndef FMUS_WINDOZE
      usleep(10);
#else
      Sleep(1); /* 1 ms */
#endif
    }
  }
  return FMUS_E_OK;
}

// pass in a filename and it will make a directory
// to contain that file
fmus_err_t mkdir_for_file (fmus_t *fmus, char *filename) {
  char *dirname = NULL;
  fmus_err_t ecode = FMUS_E_OK;
  // strip filename off path
  ecode = fmus_getpath (fmus, filename, &dirname);
//printf("[%u] %s -> %s\n", ecode, filename, dirname);
  switch (ecode) {
    case FMUS_E_OK:
      // have path, make directory
      ecode = fmus_mkdir (fmus, dirname);
      FMUS_FREEMSG(dirname, "getpath buf");
      break;
    case FMUS_E_GETPATH_NOSEP:
      // no path, just filename, no mkdir needed
      ecode = FMUS_E_OK;
      break;
    default:
      // error
      return ecode;
  }
  return ecode;
}

fmus_err_t fmus_getpath (fmus_t *fmus, char *path, char **output) {
  // *output must be freed by the caller if there is no error
  size_t j;
  size_t pathlen=0;
  u8_t nuke=0;
  if (!output) {
    FMUS_BPRINTF(fmus->tw, "BUG: fmus_getpath(output==NULL)\n");
    return FMUS_E_NULL;
  }
  *output = NULL;
  if (!path) {
    FMUS_BPRINTF(fmus->tw, "BUG: fmus_getpath(path==NULL)\n");
    return FMUS_E_NULL;
  }
  pathlen = strlen(path);
  if (!pathlen) {
    FMUS_BPRINTF(fmus->tw, "BUG: fmus_getpath(\"\")\n");
    return FMUS_E_GETPATH_ZL;
  }
  *output = FMUS_MALLOCMSG(1 + pathlen, "getpath buf");
  memcpy(*output, path, 1 + pathlen);
  // find last slash
  for (j=pathlen;j>0;j--) { // goes from L to 1, not (L-1) to 0
    nuke = ((*output)[j-1] == '/'); // is it a pathsep?
#ifdef FMUS_WINDOZE
    if ((*output)[j-1] == '\\')
      nuke = 1;
#endif
    //if ((*output)[j-1] == FMUS_PATHSEP) {
    if (nuke) {
      // got it, nuke it!
      (*output)[j-1] = 0; // null terminator
      break;
    }
  }
  if (!j) {
    // no slashes found -- fail
    FMUS_FREEMSG(*output, "getpath buf");
    *output = NULL;
    return FMUS_E_GETPATH_NOSEP;
  }
  return FMUS_E_OK;
}

static fmus_err_t fmus_mkdir(fmus_t *fmus, char *path) {

  // FIXME: probably not secure

  // recursively build a directory path
  // start at the beginning of the path and for each fragment:
  // - does the fragment exist on disc? if it does, make sure
  //   it is a directory and not a file
  // - if it doesn't, create the directory and carry on
  size_t i;
  size_t len;
  fmus_stat_t sb;
  size_t last_pathsep_index=0;
  
  char *path_clone = NULL;
  fmus_err_t ecode = FMUS_E_OK;
  
  if (!path) {
    FMUS_EPRINTF(fmus->tw, "Error: fmus_mkdir(NULL)\n");
    return FMUS_E_NULL;
  }

#ifdef FMUS_DBG
  printf("fmus_mkdir(%s)\n", path);
#endif
  
  len = strlen(path);
  path_clone = FMUS_MALLOCMSG(len+1, "fmus_mkdir path clone");
  memcpy(path_clone, path, len+1);
  
  // get rid of any multiple slashes (UNIX only, don't bother on windows)
#ifndef FMUS_WINDOZE
  for (i=0;i<len+1;i++) {
    size_t j,k;
    j=i;
    while (j<len && (path_clone[j] == FMUS_PATHSEP))
      j++;
    if (j >= len)
      j = len - 1;
    k=i;
    if (k >= len - 1)
      k = len - 2;
    if (j!=k) {
      memmove(path_clone+k+1, path_clone+j, len+1-j);
      path_clone[k+1+len-j] = 0; // null-terminate
    }
  }
  
#ifdef FMUS_DBG
  printf("fmus_mkdir path after demultislash: %s\n", path_clone);
#endif
#endif
  
  for (i=0;i<len+1;i++) {

    //size_t j=0;
    u8_t is_pathsep=0;
    char *not_a_directory_fmtstg = "";
    
    //if (path_clone[i] == FMUS_PATHSEP || !path_clone[i]) {
    is_pathsep = (path_clone[i] == '/');
#ifdef FMUS_WINDOZE
    if (path_clone[i] == '\\')
      is_pathsep = 1;
#endif
    if (is_pathsep || !path_clone[i]) {

#ifdef FMUS_WINDOZE
      // is it the drive letter?
      if (i == 2 && isalpha(path_clone[0]) && path_clone[1] == ':') {
        last_pathsep_index = i;
        continue;
      }
#endif
      
      if (!i) // leading slash, don't try to create the root directory
        continue;
      
      path_clone[i] = 0; // edit string: insert null terminator

      // OK so this is complicated slightly by the fact that
      // we can have paths like ./../../././beans/../oats/./ etc
      // but basically we never try to create a '.' or a '..' so
      // we need to detect those.

      if ((last_pathsep_index+1 < len)
          && ((!strcmp(path_clone+last_pathsep_index+1, "."))
              || !strcmp(path_clone+last_pathsep_index+1, ".."))) {
        // don't try to create these
        last_pathsep_index = i;
#ifdef FMUS_DBG
        printf("fmus_mkdir: will not create %s\n", path_clone);
#endif
        path_clone[i] = FMUS_PATHSEP; // remove null terminator again
        continue;
      }
        
      // check whether it exists
      //memset(&sb, 0, sizeof(fmus_stat_t));
      if (FMUS_E_OK == fmus_stat (fmus, path_clone, &sb)) {

//printf("fmus_stat(): path_clone=%s\n", path_clone);

        // it exists
        //if (!(statbuf.st_mode & S_IFDIR) && !(statbuf.st_mode & symlink_bit)) {
        if ((!sb.is_dir) && (!sb.is_link)) {
          // uh oh, it's not a directory or symlink
#ifdef FMUS_WINDOZE
          not_a_directory_fmtstg = "Error: Trying to create path, but \"%s\" is not a directory "
                                   "or symbolic link.\n";
#else
          not_a_directory_fmtstg = "Error: Trying to create path, but \"%s\" is not a directory.\n";
#endif
          FMUS_EPRINTF(fmus->tw,
                       not_a_directory_fmtstg,
                       path_clone);
          ecode = FMUS_E_FILE_IN_PATH;
          goto L_fmus_mkdir_end;
        }
      } else {
        if (FMUS_E_OK != (ecode = fmus_mkdir_2 (fmus, path_clone)))
          goto L_fmus_mkdir_end;
#ifdef FMUS_DBG
        printf("Created directory: %s\n", path_clone);
#endif
      }
      
      path_clone[i] = FMUS_PATHSEP; // remove null terminator again
      last_pathsep_index = i;
    }
  }
L_fmus_mkdir_end:
  if (path_clone)
    FMUS_FREEMSG(path_clone, "fmus_mkdir path clone");
  return ecode;
}

static fmus_err_t fmus_mkdir_2 (fmus_t *fmus, char *path) {
  char *errmsg = NULL;
#ifdef FMUS_DBG
  printf("fmus_mkdir_2(%s)\n", path);
#endif
#ifdef FMUS_WINDOZE
  if (_mkdir (path)) {
#else
  if (mkdir (path, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH)) { // 755
#endif
    errmsg = FMUS_STRERROR();
    FMUS_EPRINTF(fmus->tw,
                 "Error: Could not create directory \"%s\" (%s). This may be because one "
                 "of the path fragments is actually a file, or it may be a permissions "
                 "problem.\n",
                 path, errmsg);
    return FMUS_E_MKDIR;
  }
  return FMUS_E_OK;
}

u64_t read_u64_bigendian (u8_t *buf) {
  u64_t *p;
  p = (u64_t *) buf;
  return FROM_BIG_ENDIAN_64(*p);
}

s32_t read_s32_bigendian (u8_t *buf) {
  return (s32_t) read_u32_bigendian(buf);
}

float read_float_bigendian (u8_t *buf) {
  u32_t i;
  void *p;
  i = read_u32_bigendian(buf);
  p = &i;
  return *((float *) (p));
}

u32_t read_u32_bigendian (u8_t *buf) {
  u32_t *p;
  p = (u32_t *) buf;
  return FROM_BIG_ENDIAN_32(*p);
}

u16_t read_u16_bigendian (u8_t *buf) {
  u16_t *p;
  p = (u16_t *) buf;
  return FROM_BIG_ENDIAN_16(*p);
}

void write_u16_bigendian(u8_t *buf, u16_t x) {
  u16_t i = FROM_BIG_ENDIAN_16(x);
  memcpy(buf, &i, 2);
}

void write_u32_bigendian(u8_t *buf, u32_t x) {
  u32_t i = FROM_BIG_ENDIAN_32(x);
  memcpy(buf, &i, 4);
}

void write_s32_bigendian(u8_t *buf, s32_t x) {
  write_u32_bigendian(buf, (u32_t) x);
}

void write_u64_bigendian(u8_t *buf, u64_t x) {
  u64_t i = FROM_BIG_ENDIAN_64(x);
  memcpy(buf, &i, 8);
}

