#include "mpfont.h"
#include "fmus_io.h"
#include "fmus.h"
#include "cli.h"
#include <math.h>

static fmus_err_t parse_mpfont_glyphs (fmus_t *fmus,
                                       fmus_mpfont_t *mpfont,
                                       u8_t *buf2,
                                       u8_t do_kerning,
                                       size_t len,
                                       size_t minlen);
static void export_kerning_table (u8_t *buf, fmus_kern_t *kerning, u32_t len, u8_t skip_zero_adjust);

fmus_err_t fmus_load_mpfont (fmus_t *fmus, char *filename, u8_t load_kerning) {

  fmus_err_t e;
  u8_t *buf;
  size_t len;

  e = fmus_read_file(fmus,
                     filename,
                     &buf,
                     &len);
  if (e != FMUS_E_OK)
    return e;
    
  e = fmus_parse_mpfont (fmus,
                         load_kerning,
                         buf,
                         len);
                         
  FMUS_FREEMSG(buf, "fmus_read_file payload");
                         
  return e;
}

#define EPFF "Error parsing .FONT file: "

#define FMUS_MPF_MINLEN_32  0x33 // header len + zero glyphs + num kerning entries
#define FMUS_MPF_MINLEN_64  0x37 // + longer txid
// arbitrary sane limits:
#define MPF_MAX_LINE_HEIGHT         65535
#define MPF_MAX_VERT_OFFSET         65535
#define MPF_MAX_LINE_MARGIN         65535
#define MPF_MAX_FONT_SIZE           65535
#define MPF_MAX_NUM_GLYPHS          65535
#define MPF_GLYPH_MAX_LSB           127
#define MPF_GLYPH_MAX_RSB           127
#define MPF_GLYPH_MIN_LSB           -128
#define MPF_GLYPH_MIN_RSB           -128
#define MPF_GLYPH_MAX_LAYOUT_WIDTH  255
#define MPF_GLYPH_MAX_DISPLAY_WIDTH 255
#define MPF_GLYPH_MAX_HEIGHT        255
#define MPF_GLYPH_MAX_VERTOFF       255
#define MPF_GLYPH_MAX_KERNIX        255

fmus_err_t fmus_parse_mpfont (fmus_t *fmus, u8_t do_kerning, u8_t *buf, size_t len) {
  
  fmus_err_t e = FMUS_E_OK;
  fmus_mpfont_t *mpfont;
  size_t namelen;
  u8_t *buf2;
  //u32_t i;
  size_t minlen;
  
  minlen = FMUS_MPF_MINLEN_32;
  
  if (len < minlen) {
    FMUS_EPRINTF(fmus->tw, EPFF "File is too short (minimum %u bytes.)\n", minlen);
    return FMUS_E_MPF_TRUNC;
  }
  
  if (0 != memcmp(buf, "FONT", 4)) {
    FMUS_EPRINTF(fmus->tw,
                 EPFF "Bad FONT magic (0x%x 0x%x 0x%x 0x%x)\n",
                 buf[0], buf[1], buf[2], buf[3]);
    return FMUS_E_MPF_MAGIC;
  }
  
  mpfont = &(fmus->mpfont);
  
  mpfont->version = read_u32_bigendian(buf + 4);
  if ((mpfont->version != 2) && (mpfont->version != 4) && (mpfont->version != 5)) {
    FMUS_EPRINTF(fmus->tw, EPFF "This .FONT file version (%u) is not supported.\n",
                 mpfont->version);
    return FMUS_E_MPF_VERSION;
  }
  
  // check length again
  if (mpfont->version == 5) {
    minlen = FMUS_MPF_MINLEN_64;
    if (len < minlen) {
      FMUS_EPRINTF(fmus->tw, EPFF "File is too short (minimum %lu bytes.)\n", (u64_t) FMUS_MPF_MINLEN_64);
      return FMUS_E_MPF_TRUNC;
    }
  }
  
  mpfont->unk1_32 = read_u32_bigendian(buf + 8);
  
  mpfont->line_height = read_u32_bigendian(buf + 12);
  if (mpfont->line_height > MPF_MAX_LINE_HEIGHT) {
    FMUS_EPRINTF(fmus->tw, EPFF "Bad line height in .FONT file (%u, maximum %u).\n",
                 mpfont->line_height, MPF_MAX_LINE_HEIGHT);
    return FMUS_E_MPF_LINE_HEIGHT;
  }
  mpfont->vert_offset = read_u32_bigendian(buf + 16);
  if (mpfont->vert_offset > MPF_MAX_VERT_OFFSET) {
    FMUS_EPRINTF(fmus->tw, EPFF "Bad vertical offset in .FONT file (%u, maximum %u).\n",
                 mpfont->vert_offset, MPF_MAX_VERT_OFFSET);
    return FMUS_E_MPF_VERT_OFFSET;
  }
  mpfont->line_margin = read_u32_bigendian(buf + 20);
  if (mpfont->line_margin > MPF_MAX_LINE_MARGIN) {
    FMUS_EPRINTF(fmus->tw, EPFF "Bad line margin in .FONT file (%u, maximum %u).\n",
                 mpfont->line_margin, MPF_MAX_LINE_MARGIN);
    return FMUS_E_MPF_LINE_MARGIN;
  }

  mpfont->unk2_8 = read_u32_bigendian(buf + 24);
  mpfont->unk3_8 = read_u32_bigendian(buf + 25);
  mpfont->unk4_32 = read_u32_bigendian(buf + 26);
  
  mpfont->font_size = (float) read_u32_bigendian(buf + 30);
  if (mpfont->font_size > MPF_MAX_FONT_SIZE) {
    FMUS_EPRINTF(fmus->tw, EPFF "Bad font size in .FONT file (%u, maximum %u).\n",
                 mpfont->font_size, MPF_MAX_FONT_SIZE);
    return FMUS_E_MPF_FONT_SIZE;
  }
  
//printf("FONT SIZE %u\n", mpfont->font_size);  

  namelen = strlen ((char *) (buf + 34));
  if (namelen == 0)
    FMUS_WPRINTF(fmus->tw, "Warning: FONT has no name.\n");
  
  // check length again now we know the name
  minlen += namelen;
  if (len < minlen) {
    FMUS_EPRINTF(fmus->tw, EPFF "File was truncated after the header.\n");
    return FMUS_E_MPF_TRUNC;
  }

  if (mpfont->name != NULL)
    FMUS_FREEMSG(mpfont->name, "mpfont->name");
  mpfont->name = FMUS_MALLOCMSG(namelen + 1, "mpfont->name");
  memcpy (mpfont->name, buf + 34, namelen + 1);
  buf2 = buf + 34 + namelen + 1;
  
  if (mpfont->version < 5) {
    mpfont->txid_64 = read_u32_bigendian(buf2);
  } else {
    mpfont->txid_64 = read_u64_bigendian(buf2);
    buf2 += 4;
  }
    
//printf("txid_64=%llx\n", mpfont->txid_64);
    
  mpfont->tx_fmt = read_u32_bigendian(buf2 + 4);
  if ((mpfont->tx_fmt > 4) && (mpfont->tx_fmt != 8)) {
    FMUS_EPRINTF (fmus->tw, EPFF ".FONT file has an unknown texture format value (%u).\n", mpfont->tx_fmt);
    return FMUS_E_MPF_TX_FMT;
  }
  mpfont->num_glyphs = read_u32_bigendian(buf2 + 8);
  if (mpfont->num_glyphs > MPF_MAX_NUM_GLYPHS) {
    FMUS_EPRINTF(fmus->tw, EPFF ".FONT file contains too many glyphs (%u, maximum %u).\n",
                 mpfont->num_glyphs, MPF_MAX_NUM_GLYPHS);
    return FMUS_E_MPF_NUM_GLYPHS;
  } else if ((mpfont->num_glyphs == 0)
             && (!fmus->edit_mode)
             && (!fmus->do_not_load)
             && (!fmus->query_only)) {
    FMUS_WPRINTF(fmus->tw, ".FONT file contains no glyphs! Output texture will be blank!\n");
  }
  
//printf("num_glyphs=%u\n", mpfont->num_glyphs);
  
  buf2 += 12;

  // get glyphs
  if (mpfont->num_glyphs > 0) {
//printf("allocing glyphs [2]\n");
    fmus->glyphs = FMUS_MALLOCMSG(sizeof(fmus_glyph_t) * mpfont->num_glyphs, "fmus->glyphs");
    e = parse_mpfont_glyphs (fmus, &(fmus->mpfont), buf2, do_kerning, len, minlen);
    if (e != FMUS_E_OK) {
      FMUS_FREEMSG(fmus->glyphs, "fmus->glyphs");
      fmus->glyphs=NULL;
      return e;
    }
  }
  
  print_mpfont(fmus, mpfont, fmus->mpfont_filename, fmus->verbose_query);

  //FMUS_FREEMSG(fmus->glyphs, "fmus->glyphs");
  //fmus->glyphs=NULL;

  return FMUS_E_OK;
  
}

static fmus_err_t parse_mpfont_glyphs (fmus_t *fmus,
                                       fmus_mpfont_t *mpfont,
                                       u8_t *buf2,
                                       u8_t do_kerning,
                                       size_t len,
                                       size_t minlen) {
  
  u32_t i;
  u32_t kerning_count;
  
//printf("num_glyphs=%u\n", mpfont->num_glyphs);
  
  for (i=0; i < mpfont->num_glyphs; i++) {
    
    fmus_glyph_t *glyph;
    
    if (mpfont->version == 2)
      minlen += 0x2e; // glyph len
    else
      minlen += 0x1b; // glyph len
    
    // check length again
    if (len < minlen) {
      FMUS_EPRINTF(fmus->tw, EPFF "File was truncated during glyph table.\n");
      return FMUS_E_MPF_TRUNC;
    }
    
    glyph = fmus->glyphs + i;
    glyph->utf16 = read_u16_bigendian(buf2);
    
    glyph->tx_left = read_float_bigendian(buf2 + 2);
    if ((glyph->tx_left < 0.0) || (glyph->tx_left > 1.0)) {
      FMUS_EPRINTF(fmus->tw, EPFF "Glyph %u: bad left texture coordinate (0.0 <= %f <= 1.0)\n", i, glyph->tx_left);
      return FMUS_E_MPF_TX_COORD;
    }
    glyph->tx_top = read_float_bigendian(buf2 + 6);
    if ((glyph->tx_top < 0.0) || (glyph->tx_top > 1.0)) {
      FMUS_EPRINTF(fmus->tw, EPFF "Glyph %u: bad top texture coordinate (0.0 <= %f <= 1.0)\n", i, glyph->tx_top);
      return FMUS_E_MPF_TX_COORD;
    }
    glyph->tx_right = read_float_bigendian(buf2 + 10);
    if ((glyph->tx_right < 0.0) || (glyph->tx_right > 1.0)) {
      FMUS_EPRINTF(fmus->tw, EPFF "Glyph %u: bad right texture coordinate (0.0 <= %f <= 1.0)\n", i, glyph->tx_right);
      return FMUS_E_MPF_TX_COORD;
    }
    glyph->tx_bottom = read_float_bigendian(buf2 + 14);
    if ((glyph->tx_bottom < 0.0) || (glyph->tx_bottom > 1.0)) {
      FMUS_EPRINTF(fmus->tw, EPFF "Glyph %u: bad bottom texture coordinate (0.0 <= %f <= 1.0)\n", i, glyph->tx_bottom);
      return FMUS_E_MPF_TX_COORD;
    }
    
//printf("L %f T %f R %f B %f\n", glyph->tx_left, glyph->tx_top, glyph->tx_right, glyph->tx_bottom);
    
    //buf2 += 18;
    
    if (mpfont->version >= 4) {
      // extra field in v4/5: RGBA channel ix
      glyph->rgba_chnl_ix = buf2[18];
      glyph->lsb = (s8_t) buf2[19];
      glyph->layout_width = buf2[20];
      glyph->rsb = (s8_t) buf2[21];
      glyph->display_width = buf2[22];
      glyph->height = buf2[23];
      glyph->vert_offset = buf2[24];
      glyph->kern_ix = read_u16_bigendian(buf2 + 25);
    } else {
      glyph->lsb = read_s32_bigendian (buf2 + 18);
      glyph->layout_width = read_u32_bigendian (buf2 + 22);
      glyph->rsb = read_s32_bigendian (buf2 + 26);
      glyph->display_width = read_u32_bigendian (buf2 + 30);
      glyph->height = read_u32_bigendian (buf2 + 34);
      glyph->vert_offset = read_u32_bigendian (buf2 + 38);
      glyph->kern_ix = read_u32_bigendian (buf2 + 42);
    }

    if ((glyph->lsb < MPF_GLYPH_MIN_LSB) || (glyph->lsb > MPF_GLYPH_MAX_LSB)) {
      FMUS_EPRINTF(fmus->tw,
                   EPFF "Glyph %u: bad left side bearing value (%u -- value must lie between %u and %u).\n",
                   i,
                   glyph->lsb,
                   MPF_GLYPH_MIN_LSB,
                   MPF_GLYPH_MAX_LSB);
      return FMUS_E_MPF_LSB;
    } else if ((glyph->rsb < MPF_GLYPH_MIN_RSB) || (glyph->rsb > MPF_GLYPH_MAX_RSB)) {
      FMUS_EPRINTF(fmus->tw,
                   EPFF "Glyph %u: bad right side bearing value (%u -- value must lie between %u and %u).\n",
                   i,
                   glyph->rsb,
                   MPF_GLYPH_MIN_RSB,
                   MPF_GLYPH_MAX_RSB);
      return FMUS_E_MPF_RSB;
    } else if (glyph->layout_width > MPF_GLYPH_MAX_LAYOUT_WIDTH) {
      FMUS_EPRINTF(fmus->tw,
                   EPFF "Glyph %u: bad layout width (%u, maximum %u).\n",
                   i,
                   glyph->layout_width,
                   MPF_GLYPH_MAX_LAYOUT_WIDTH);
      return FMUS_E_MPF_LAYOUT_WIDTH;
    } else if (glyph->display_width > MPF_GLYPH_MAX_DISPLAY_WIDTH) {
      FMUS_EPRINTF(fmus->tw,
                   EPFF "Glyph %u: bad display width (%u, maximum %u).\n",
                   i,
                   glyph->display_width,
                   MPF_GLYPH_MAX_DISPLAY_WIDTH);
      return FMUS_E_MPF_WIDTH;
    } else if (glyph->height > MPF_GLYPH_MAX_HEIGHT) {
      FMUS_EPRINTF(fmus->tw, EPFF
                   "Glyph %u: bad height (%u, maximum %u).\n",
                   i,
                   glyph->height,
                   MPF_GLYPH_MAX_HEIGHT);
      return FMUS_E_MPF_HEIGHT;
    } else if (glyph->vert_offset > MPF_GLYPH_MAX_VERTOFF) {
      FMUS_EPRINTF(fmus->tw,
                   EPFF "Glyph %u: bad vertical offset (%u, maximum %u).\n",
                   i,
                   glyph->vert_offset,
                   MPF_GLYPH_MAX_VERTOFF);
      return FMUS_E_MPF_GLYPH_VERT_OFFSET;
    } else if (glyph->kern_ix > MPF_GLYPH_MAX_KERNIX) {
      FMUS_EPRINTF(fmus->tw,
                   EPFF "Glyph %u: bad kerning index (%u, maximum %u).\n",
                   i,
                   glyph->kern_ix,
                   MPF_GLYPH_MAX_KERNIX);
      return FMUS_E_MPF_KERN_IX;
    }
    
//printf("%f %f %f %f\n", glyph->tx_left, glyph->tx_right, glyph->tx_top, glyph->tx_bottom);
    if (mpfont->version == 2)
      buf2 = buf2 + 0x2e;
    else
      buf2 = buf2 + 0x1b;
    
  }
  
  //FMUS_MPRINTF(fmus->tw, "FONT %s loaded with %u glyphs.\n", fmus->mpfont_filename, mpfont->num_glyphs);

  //kerning_count = (u32_t) read_u32_bigendian (buf2);
  
#define KERNING_LEN_MAX 0xffff

//print_mpfont(fmus, &(fmus->mpfont), "whatever", 1);

  if (do_kerning) {
    //minlen+=4;
    // check length again
    //if (len < minlen) {
    //  FMUS_EPRINTF(fmus->tw, EPFF "File was truncated during kerning count value.\n");
    //  return FMUS_E_MPF_TRUNC;
   // }
    //buf2 += 4;
    kerning_count = (u32_t) read_u32_bigendian (buf2);
    if (kerning_count > KERNING_LEN_MAX) {
      FMUS_EPRINTF(fmus->tw, EPFF "Bad kerning table length value: %x (max. %u)\n",
                   kerning_count, KERNING_LEN_MAX);
      return FMUS_E_MPF_KERNLEN;
    }
//printf("kc=%u\n", kerning_count);
    if (fmus->kerning != NULL)
      FMUS_FREEMSG(fmus->kerning, "fmus->kerning");
    fmus->kerning = NULL;
    if (kerning_count)
      fmus->kerning = FMUS_MALLOCMSG (kerning_count * sizeof(fmus_kern_t), "fmus->kerning");
    buf2 += 4;
    for (i=0;i<kerning_count;i++) {
      minlen += 8;
      if (len < minlen) {
        FMUS_EPRINTF(fmus->tw,
                     EPFF "File was truncated during kerning table (entry %u of %u expected).\n",
                     i, kerning_count);
        FMUS_FREEMSG(fmus->kerning, "fmus->kerning");
        fmus->kerning = NULL;
        return FMUS_E_MPF_TRUNC;
      }
      fmus->kerning[i].c1_utf16 = read_u16_bigendian(buf2);
      fmus->kerning[i].c2_utf16 = read_u16_bigendian(buf2 + 2);
      fmus->kerning[i].adjust = read_s32_bigendian(buf2 + 4);
      buf2 += 8;
    }
    mpfont->kerning_entries = kerning_count;
//printf("kc=%u\n", kerning_count);
  }
  

  return FMUS_E_OK;

}

fmus_err_t export_mpfont(fmus_t *fmus,
                         fmus_mpfont_t *mpfont,
                         u8_t do_kerning,
                         u8_t **font_out,
                         size_t *font_len_out) {
  u8_t glyph_record_len;
  u8_t *buf;
  u32_t i;
  if ((mpfont->version != 2) && (mpfont->version != 4) && (mpfont->version != 5)) {
    FMUS_EPRINTF(fmus->tw, "Error exporting font: BUG: bad .FONT version: %x\n", mpfont->version);
    return FMUS_E_BUG;
  }
  if (mpfont->version == 2)
    glyph_record_len = 46; //0x2e;
  else
    glyph_record_len = 0x1b;
  // predict size
  *font_len_out = 50
                  + ((mpfont->version == 5) ? 4 : 0) // 64 bit txid?
                  + (glyph_record_len * mpfont->num_glyphs)
                  + strlen(mpfont->name) + 1
                  + (8 * mpfont->kerning_entries);
  //FMUS_MPRINTF(fmus->tw, "Exported .FONT will be %lu bytes.\n", *font_len_out);
  *font_out = FMUS_MALLOCMSG(*font_len_out, "FONT");
  memcpy(*font_out, "FONT", 4);
  write_u32_bigendian(*font_out + 4,  mpfont->version);
  write_u32_bigendian(*font_out + 8, mpfont->unk1_32);  // unk1
  write_u32_bigendian(*font_out + 12, mpfont->line_height); // line height (i.e. on carriage return)
  write_u32_bigendian(*font_out + 16, mpfont->vert_offset); // vertical offset
  write_u32_bigendian(*font_out + 20, mpfont->line_margin); // line margin
  (*font_out)[24] = mpfont->unk2_8; // unk2
  (*font_out)[25] = mpfont->unk3_8; // unk3
  write_u32_bigendian(*font_out + 26, mpfont->unk4_32); // unk4
  // FIXME: use real pixel font size value that was given to freetype?
  // this field doesn't seem to be used by the game anyway (only 0-00 checked tho)
//printf("font_size=%u\n", mpfont->font_size);
  write_u32_bigendian(*font_out + 30, (s32_t) roundf(mpfont->font_size)); // font size
  memcpy(*font_out + 34, mpfont->name, strlen(mpfont->name) + 1);
  buf = *font_out + 34 + strlen(mpfont->name) + 1;
  if (mpfont->version < 5)
    write_u32_bigendian(buf, mpfont->txid_64 & 0xffffffff); // tx ID
  else {
    write_u64_bigendian(buf, mpfont->txid_64); // tx ID
    buf += 4;
  }
  write_u32_bigendian(buf + 4, mpfont->tx_fmt); // tx format
  write_u32_bigendian(buf + 8, mpfont->num_glyphs); // glyph count
  buf += 12;
  
  for (i=0;i<mpfont->num_glyphs;i++) {
    
    fmus_glyph_t *glyph;
    union {
      float f;
      u32_t i;
    } u;
    glyph = fmus->glyphs + i;
    
    if (!do_kerning)
      glyph->kern_ix = 0; // off the end of the kerning table => no kerning
        
    write_u16_bigendian(buf, glyph->utf16);
    
    // union shenanigans
    u.f = glyph->tx_left;
    write_u32_bigendian(buf + 2, u.i);
    u.f = glyph->tx_top;
    write_u32_bigendian(buf + 6, u.i);
    u.f = glyph->tx_right;
    write_u32_bigendian(buf + 10, u.i);
    u.f = glyph->tx_bottom;
    write_u32_bigendian(buf + 14, u.i);
    
//printf("L %f T %f R %f B %f\n", glyph->tx_left, glyph->tx_top, glyph->tx_right, glyph->tx_bottom);

//printf("kern_ix=%u\n", glyph->kern_ix);
    
    if (mpfont->version == 2) {
      write_s32_bigendian(buf + 18, glyph->lsb);
      write_u32_bigendian(buf + 22, glyph->layout_width);
      write_s32_bigendian(buf + 26, glyph->rsb);
      write_u32_bigendian(buf + 30, glyph->display_width);
      write_u32_bigendian(buf + 34, glyph->height);
      write_u32_bigendian(buf + 38, glyph->vert_offset);
      write_u32_bigendian(buf + 42, glyph->kern_ix);
    } else {
      buf[18] = glyph->rgba_chnl_ix;
      buf[19] = 0xff & glyph->lsb;
      buf[20] = 0xff & glyph->layout_width;
      buf[21] = 0xff & glyph->rsb;
      buf[22] = 0xff & glyph->display_width;
      buf[23] = 0xff & glyph->height;
      buf[24] = 0xff & glyph->vert_offset;
      write_u16_bigendian(buf + 25, glyph->kern_ix);
    }
    buf += glyph_record_len;
  }
  
  export_kerning_table (buf, fmus->kerning, mpfont->kerning_entries, 0); //fmus->verbatim_copy);
  
  return FMUS_E_OK;
  
}

static void export_kerning_table (u8_t *buf, fmus_kern_t *kerning, u32_t len, u8_t skip_zero_adjust) {
  u32_t i, j;
  u32_t newlen;
  newlen=len;
  if (skip_zero_adjust)
    for (i=0; i<len; i++)
      if (kerning[i].adjust == 0)
        newlen--;
  write_u32_bigendian(buf, newlen); // kerning table length
  j=0;
  for (i=0; i<len; i++) {
    if (skip_zero_adjust && (kerning[i].adjust == 0))
      continue;
    write_u16_bigendian(buf+4, kerning[i].c1_utf16);
    write_u16_bigendian(buf+6, kerning[i].c2_utf16);
    write_s32_bigendian(buf+8, kerning[i].adjust);
    j++;
    buf += 8;
  }
}

void init_mpfont (fmus_mpfont_t *mpfont) {

  char *default_font_name = "Untitled" WATERMARK;
  size_t namelen;
  
  memset(mpfont, 0, sizeof(fmus_mpfont_t));

  mpfont->version = 2;
  mpfont->line_height = 33;
  mpfont->vert_offset = 22;
  mpfont->font_size = 18;
  mpfont->unk1_32 = 0x17;
  mpfont->unk4_32 = 0x7;
  
  namelen = strlen(default_font_name);
  mpfont->name = FMUS_MALLOCMSG(namelen+1, "mpfont->name");
  memcpy(mpfont->name, default_font_name, namelen+1);
  
}

static const char *mpfont_tx_fmts[9] = {
  "fill on L0",
  "fill on L0; outline on L1",
  "fill on L0, L1, L2 & L3",
  "fill on L0 & L2; outline on L1 & L3",
  "fill on L0 & L1; copies on L2 & L3",
  NULL,
  NULL,
  NULL,
  "fill on L2 & L3, outline on L0 & L1"
};

void print_mpfont(fmus_t *fmus,
                  fmus_mpfont_t *mpfont,
                  char *mpfont_filename,
                  u8_t verbose) {
                    
  u32_t i;
                    
  if (NULL != mpfont_filename)
    FMUS_MPRINTF(fmus->tw, "%s:\n", mpfont_filename);
  FMUS_MPRINTF(fmus->tw, "  \"%s\"\n", mpfont->name);
  if (mpfont->version == 2)
    FMUS_MPRINTF(fmus->tw, "    - Version:        2 (Metroid Prime NTSC \"0-00\")\n");
  else if (mpfont->version == 4)
    FMUS_MPRINTF(fmus->tw, "    - Version:        4 (Metroid Prime PAL; Metroid Prime 2; MP1/MP2 Trilogy)\n");
  else if (mpfont->version == 5)
    FMUS_MPRINTF(fmus->tw, "    - Version:        5 (Metroid Prime 3; DKCR)\n");
  else
    FMUS_MPRINTF(fmus->tw, "    - Version:        %u (?)\n", mpfont->version);
  if (verbose)
    FMUS_MPRINTF(fmus->tw,   "    - Unknown 1 (32): 0x%x\n", mpfont->unk1_32);
  FMUS_MPRINTF(fmus->tw, "    - Line height:    %u\n", mpfont->line_height);
  FMUS_MPRINTF(fmus->tw, "    - Vert. offset:   %u\n", mpfont->vert_offset);
  FMUS_MPRINTF(fmus->tw, "    - Line margin:    %u\n", mpfont->line_margin);
  if (verbose) {
    FMUS_MPRINTF(fmus->tw,   "    - Unknown 2 (8):  0x%x\n", mpfont->unk2_8);
    FMUS_MPRINTF(fmus->tw,   "    - Unknown 3 (8):  0x%x\n", mpfont->unk3_8);
    FMUS_MPRINTF(fmus->tw,   "    - Unknown 4 (32): 0x%x\n", mpfont->unk4_32);
  }
  FMUS_MPRINTF(fmus->tw, "    - Font size:      %u\n", mpfont->font_size);
  if (mpfont->version < 5)
    FMUS_MPRINTF(fmus->tw, "    - Texture:        0x%x\n", (u32_t) (mpfont->txid_64 & 0xffffffff));
  else
    FMUS_MPRINTF(fmus->tw, "    - Texture:        0x%lx\n", mpfont->txid_64);
  FMUS_MPRINTF(fmus->tw, "    - Texture format: 0x%x (%s)\n", mpfont->tx_fmt, mpfont_tx_fmts[mpfont->tx_fmt]);
  FMUS_MPRINTF(fmus->tw, "    - Glyphs:         %u\n", mpfont->num_glyphs);
  FMUS_MPRINTF(fmus->tw, "    - Kern. entries:  %u\n", mpfont->kerning_entries);
  
  if (verbose && (fmus->glyphs!=NULL) && mpfont->num_glyphs) {
    // print glyphs
    for (i=0; i<mpfont->num_glyphs; i++) {
      printf("G%5u: ", i);
      if (fmus->glyphs + i == NULL) {
        printf("NULL\n");
        continue;
      }
      printf("Ux%04x ", fmus->glyphs[i].utf16); // 6
      printf("%s=%.4f ", glyph_edit_labels[GELIX_TXCL], fmus->glyphs[i].tx_left);
      printf("%s=%.4f ", glyph_edit_labels[GELIX_TXCT], fmus->glyphs[i].tx_top);
      printf("%s=%.4f ", glyph_edit_labels[GELIX_TXCR], fmus->glyphs[i].tx_right);
      printf("%s=%.4f ", glyph_edit_labels[GELIX_TXCB], fmus->glyphs[i].tx_bottom);
      printf("chnl=%u ", fmus->glyphs[i].rgba_chnl_ix);
      printf("%s=%03i\n", glyph_edit_labels[GELIX_LSB], fmus->glyphs[i].lsb);
      printf("        %s=%03u ", glyph_edit_labels[GELIX_LWIDTH], fmus->glyphs[i].layout_width);
      printf("%s=%03i ", glyph_edit_labels[GELIX_RSB], fmus->glyphs[i].rsb);
      printf("%s=%03u ", glyph_edit_labels[GELIX_DWIDTH], fmus->glyphs[i].display_width);
      printf("%s=%03u ", glyph_edit_labels[GELIX_HEIGHT], fmus->glyphs[i].height); // height
      printf("%s=%03u ", glyph_edit_labels[GELIX_VERTOFF], fmus->glyphs[i].vert_offset); // vert offset
      printf("kernix=%u\n", fmus->glyphs[i].kern_ix); // kerning ix
    }
  }
  
}
