#include "text.h"
#include "fmus.h"
#include "fmus_io.h"
#include <math.h>

// The caller must free/destroy bitmap_out.
fmus_err_t fmus_text_render (fmus_t *fmus,
                             fmus_text_t *text,
                             s32_t face_ix,
                             u32_t utf32,
                             float px_size,
                             s32_t *width_out,
                             s32_t *height_out,
                             u8_t **bitmap_out,
                             s32_t *baseline_y_out,
                             s32_t *advance_out,
                             s32_t *left_bearing_out) {
  
  fmus_err_t e;
  int ft_err;
  FT_Face face;
  u8_t pass;
  FT_GlyphSlot slot;
  s32_t bitmap_top;
  s32_t bitmap_bottom;
  s32_t bitmap_rows_s32;
  s32_t display_width;
  s32_t output_w;
  s32_t output_h;
  s32_t wxh;
  
  if (NULL == bitmap_out) {
    FMUS_EPRINTF(fmus->tw, "fmus_text_render: BUG: bitmap_out is NULL, nothing to do.\n");
    return FMUS_E_NULL;
  }
  
  e = FMUS_E_OK;
  *bitmap_out = NULL;
  output_w = 0;
  output_h = 0;
  bitmap_top = 0;
  bitmap_bottom = 0;
  bitmap_rows_s32 = 0;
  face = text->font.faces[face_ix];
  slot = face->glyph;
  
  e = fmus_text_set_pixel_size(fmus, text, px_size, face_ix);
  if (e != FMUS_E_OK)
    return e;
  
  for (pass=0; pass < 2; pass++) { // on first pass, just calculate the size
    
    size_t src_ix; // for copying bitmap
    size_t dst_ix;
    size_t src_ix_orig;
    s32_t x; // also for copying bitmap
    s32_t y;
    FT_UInt glyph_index;
    s32_t pen_x;
    s32_t pen_y;
    FT_Bitmap ft_bitmap_intermediate;
    FT_Bitmap *ft_bitmap;
    
//printf("pass %u\n", pass);
    
    *baseline_y_out = 0;    
    pen_x = 0;
    if (!pass)
      pen_y = 0;
    else {
      pen_y = bitmap_top;
      *baseline_y_out = bitmap_top;
    }
    
    // retrieve glyph index from character code
    // FIXME: move this outside the pass loop ...
    glyph_index = FT_Get_Char_Index(face, utf32);
    
    if (glyph_index == 0) {
      //if (print_warnings) {
      //  FMUS_WPRINTF(fmus->tw,
      //               "Warning: Font does not contain a glyph for Unicode 0x%x.\n",
      //               utf32);
      //}
      //continue;
      e = FMUS_E_FT_GET_CHAR_IX;
      goto Lfail;
    }

    // load glyph image into the slot (erase previous one)
    // FIXME: we're using two passes, so don't do this twice;
    // find a way of storing it after the first pass and
    // using it again for the second
    ft_err = FT_Load_Glyph(face, glyph_index, FT_LOAD_DEFAULT);
    if (ft_err < 0) {
      FMUS_EPRINTF(fmus->tw,
                   "ERROR: fmus_text_render: failed to load glyph via FT_Load_Glyph for unicode 0x%x (code %i). Aborting.\n",
                   utf32,
                   ft_err);
      e = FMUS_E_FT_LOAD_GLYPH;
      goto Lfail;
    }
      
    // convert to a 1-bit bitmap
    // FIXME: we're using two passes, so don't do this twice;
    // find a way of storing it after the first pass and
    // using it again for the second
    ft_err = FT_Render_Glyph(face->glyph, FT_RENDER_MODE_MONO);
    
    if (ft_err < 0) {
      FMUS_EPRINTF(fmus->tw,
                   "ERROR: fmus_text_render: failed to render glyph via FT_Render_Glyph for unicode 0x%x (code %i), ignoring char\n",
                   utf32,
                   ft_err);
      e = FMUS_E_FT_RENDER_GLYPH;
      goto Lfail;
    }

//if (text->utf32_buf[i] == '_')
//printf("bm_top=%i, bm_bottom=%i, bm_rows=%i\n", slot->bitmap_top, (slot->bitmap.rows - slot->bitmap_top), slot->bitmap.rows);
    
    ft_bitmap = &(slot->bitmap);
    bitmap_rows_s32 = (0x7fffffff & slot->bitmap.rows);
    display_width = slot->bitmap.width;
    bitmap_top = slot->bitmap_top;
    bitmap_bottom = bitmap_rows_s32 - slot->bitmap_top;
      
//printf("%c: slot->bitmap.rows = %i\n", utf32, slot->bitmap.rows);

    if (pass) { // second pass
      
      // convert from 1 bpp to 8 bpp
    
      FT_Bitmap_Init(&ft_bitmap_intermediate);
      FT_Bitmap_Convert(text->library, ft_bitmap, &ft_bitmap_intermediate, 1);
   
      // at this point ft_bitmap_intermediate.buffer will have a byte per pixel representation
      // copy to destination
      
      // we need to respect any possible stride
      // on the freetype bitmap so we can't just memcpy()
      
//printf("pitch %u\n", ft_bitmap_intermediate.pitch);

memset(*bitmap_out, 0, *width_out * *height_out);
    
      src_ix=0;
      dst_ix=0;
      //width_s32 = (s32_t) (0x7fffffff & (slot->bitmap.width));
      for (y=0; y<*height_out; y++) {
        src_ix_orig = src_ix;
        for (x=0; x < *width_out; x++) {
          (*bitmap_out)[dst_ix] = ft_bitmap_intermediate.buffer[src_ix] ? 255 : 0;
//printf("%c", ft_bitmap_intermediate.buffer[src_ix]?'#':' ');
          src_ix++;
          dst_ix++;
        }
//printf("\n");
        src_ix = src_ix_orig + ft_bitmap_intermediate.pitch;
      }
 
      FT_Bitmap_Done(text->library, &ft_bitmap_intermediate);
    }
  
    // increment pen position
    pen_x += (slot->advance.x) >> 6;
    pen_y += (slot->advance.y) >> 6;
    
    if (!pass) { // end of first pass
      
      //output_w = pen_x; // - pen_initial.x; // look how far we've come
      output_w = display_width; //(slot->metrics.width /* + slot->metrics.horiBearingX*/) >> 6;
      output_h = bitmap_top + bitmap_bottom; //tallest_glyph_height; //pen.y - pen_initial.y;

      // since the outline versions of everything are wider,
      // if we're running outline mode, we 
      *width_out = output_w;
      *height_out = output_h;
      *advance_out = (slot->advance.x) >> 6;
      *left_bearing_out = (slot->metrics.horiBearingX >> 6);
      
//printf("width_out = %i, height_out=%i, lbearing_out=%i\n",
//       *width_out, *height_out, *left_bearing_out);
      
      if (output_w == 0) {
        FMUS_WPRINTF(fmus->tw, "Warning: Glyph for Unicode 0x%x has rendered width of 0.\n", utf32);
        break;
      }
      
      if (output_h == 0) {
        FMUS_WPRINTF(fmus->tw, "Warning: Glyph for Unicode 0x%x has rendered height of 0.\n", utf32);
        break;
      }
      
      wxh = output_w * output_h;
      
      *bitmap_out = FMUS_MALLOCMSG(wxh, "glyph bitmap out");
      
    }
    
  } // next pass
  
  goto Ldone;

Lfail:
  if (*bitmap_out)
    FMUS_FREEMSG(*bitmap_out, "glyph bitmap_out");
  *bitmap_out = NULL;
Ldone:
  return e;
}

void set_mpfont_name_from_freetype (fmus_text_t *text, s32_t font_face_ix, char **font_name_inout) {
  size_t family_len, style_len, new_name_len;
  FT_Face face;
  if (*font_name_inout != NULL) {
    face = text->font.faces[font_face_ix];
    family_len = strlen(face->family_name);
    style_len = strlen(face->style_name);
    new_name_len = family_len + 2 + style_len + 1;
    FMUS_FREEMSG(*font_name_inout, "mpfont->name");
    *font_name_inout = FMUS_MALLOCMSG(new_name_len + 1, "mpfont->name");
    FMUS_SNPRINTF(*font_name_inout,
                  new_name_len + 1,
                  new_name_len + 1,
                  "%s [%s]",
                  face->family_name,
                  face->style_name);
  }
}

fmus_err_t get_kerning_from_freetype(fmus_t *fmus, fmus_text_t *text, fmus_mpfont_t *mpfont, s32_t font_face_ix) {
  u32_t i;
  //fmus_kern_t *kt; // kerning table
  //size_t nke; // num kerning entries
  u8_t pass;
  int result;

  //kt = NULL;
  //nke = 0;
  
  for (pass=0; pass<2; pass++) { // FIXME: half the speed it could go
    size_t count=0;
    for (i=0; i<mpfont->num_glyphs; i++) {
      u32_t j;
      FT_Vector v;
      //FT_Kerning_Mode kerning_mode;

      for (j=0;j<mpfont->num_glyphs;j++) {
        u32_t gix1, gix2;
        gix1 = FT_Get_Char_Index(text->font.faces[font_face_ix], fmus->glyphs[i].utf16);
        gix2 = FT_Get_Char_Index(text->font.faces[font_face_ix], fmus->glyphs[j].utf16);
//printf("%c.%c\n", fmus->glyphs[i].utf16, fmus->glyphs[j].utf16);
        result = FT_Get_Kerning(text->font.faces[font_face_ix],
                                gix1,
                                gix2,
                                FT_KERNING_DEFAULT,
                                &v);
        if (result != 0)
          continue;
        if (v.x) {
          if (pass) {
            fmus->kerning[count].c1_utf16 = fmus->glyphs[i].utf16;
            fmus->kerning[count].c2_utf16 = fmus->glyphs[j].utf16;
            fmus->kerning[count].adjust = v.x;
          }
          count++;
        }
      }
    }
//printf("count=%u\n", count);
    if (!count)
      break;
    if (!pass) {
      fmus->kerning = FMUS_MALLOCMSG(sizeof(fmus_kern_t) * count, "fmus->kerning");
      mpfont->kerning_entries = count;
    }
  }
  
  return FMUS_E_OK;
}

fmus_err_t fmus_text_set_point_size (fmus_t *fmus, fmus_text_t *text, float size_pt, u32_t face_id) {
  
  fmus_err_t e;
  int ft_err;
  s32_t x;
  
  float haxx;
  
  e = FMUS_E_OK;
  
  // FIXME: check font_id, face_id are sane
  
  if (fmus->ddpi == 0.0 || fmus->hdpi == 0.0 || fmus->vdpi == 0.0) {
    FMUS_EPRINTF(fmus->tw, "fmus_text_set_point_size: BUG: dpi not set correctly\n");
    return FMUS_E_BUG;
  }
  
  // hax the DPI based on ratio between window size and fullscreen size
  //hax = ((float) fmus->cfg.windowheight) / ((float) dmode.h);
  haxx = 1.0f;
  x = (s32_t) roundf(64.0f * size_pt * haxx);

  ft_err = FT_Set_Char_Size(text->font.faces[face_id], // handle to face object
                            x,    // char_width in 1/64th of points
                            x,    // char_height in 1/64th of points
                            (FT_UInt) fmus->hdpi,  // horizontal device resolution
                            (FT_UInt) fmus->vdpi); // vertical device resolution

  if (ft_err != 0) {
    FMUS_EPRINTF(fmus->tw, "fmus_text_set_point_size: FT_Set_Char_Size failed (code %u)\n", ft_err);
    return FMUS_E_FT_SET_CHAR_SIZE;
  }

  return e;
  
}

fmus_err_t fmus_text_set_pixel_size (fmus_t *fmus, fmus_text_t *text, float size_px, u32_t face_id) {
  
  fmus_err_t e;
  int ft_err;
  float rounded;
  
  e = FMUS_E_OK;
  
  rounded = roundf(size_px);

  ft_err = FT_Set_Pixel_Sizes(text->font.faces[face_id], (FT_UInt) rounded, (FT_UInt) rounded);
                            
  if (ft_err != 0) {
    FMUS_EPRINTF(fmus->tw, "fmus_text_set_pixel_size: FT_Set_Char_Size failed (code %u)\n", ft_err);
    return FMUS_E_FT_SET_CHAR_SIZE;
  }

  return e;
  
}

fmus_err_t fmus_text_init(fmus_t *fmus, fmus_text_t *text) {
  int ft_err;
  fmus_err_t e = FMUS_E_OK;
  ft_err = FT_Init_FreeType(&(text->library));
  if (ft_err != 0) {
    FMUS_EPRINTF(fmus->tw, "Could not initialise FreeType for text rendering (code %i).\n", ft_err);
  }
  FMUS_MPRINTF(fmus->tw, "FreeType initialised.\n");
  //e = fmus_load_font(fmus, text, FONT_ID_MAIN);
  return e;
}

fmus_err_t fmus_text_load_font(fmus_t *fmus,
                               fmus_text_t *text,
                               char *fontfile,
                               s32_t desired_face_ix) {
  
  fmus_err_t e = FMUS_E_OK;
  //u8_t *payload = NULL;
  int ft_err;
  fmus_font_t *font;
  s32_t num_faces;
  s32_t face_ix;
  
  font = &(text->font);
  
  // load font file
  e = fmus_read_file (fmus, fontfile, &(font->file_contents), &(font->file_len));
  if (e != FMUS_E_OK)
    return e;
    
  //FMUS_MPRINTF(fmus->tw, "Loaded font file, %lu bytes: %s", font->file_len, fmus_font_filenames[font_id]);
  
  // load all the faces
  face_ix = 0;
  num_faces = 0;
  
  font->faces = FMUS_MALLOCMSG(sizeof(FT_Face), "font->faces"); // just one
  
  do {
    
    FT_Face face = NULL;
    ft_err = FT_New_Memory_Face (text->library, font->file_contents, font->file_len, face_ix, (font->faces) + face_ix);
    //size_t family_len, style_len, new_name_len;
    
    face = (font->faces)[face_ix];
    
    if (ft_err != 0) {
      FMUS_EPRINTF(fmus->tw, "Error loading TrueType / OpenType font (FreeType code %i): %s\n", ft_err, fontfile);
      //FMUS_FREEMSG(font->file_contents, "readfile buf");
      //font->file_contents = NULL;
      e = FMUS_E_FT_NEW_FACE;
      goto Lfail;
    }
    
    FMUS_MPRINTF(fmus->tw,
                 "  - Loaded font face %u with %u glyphs: [%s] [%s]\n",
                 face_ix,
                 face->num_glyphs,
                 face->family_name,
                 face->style_name);
    
    if (!face_ix) {
      num_faces = face->num_faces;
      FMUS_MPRINTF(fmus->tw, "  - Font contains %u %s.\n", num_faces, ((num_faces==1) ? "face" : "faces"));
      font->faces = FMUS_REALLOCMSG(font->faces, sizeof(FT_Face) * num_faces, "font->faces");
    } else if (face->num_faces != num_faces) { // check we're getting consistently the same num_faces
      FMUS_EPRINTF(fmus->tw, "BUG: fmus_load_font: FT_New_Memory_Face returns inconsistent num_faces\n");
      e = FMUS_E_BUG;
      goto Lfail;
    }
    
//printf("%u\n", face_ix);
    
    face_ix++;
    
  } while (face_ix < num_faces);
  
  if (desired_face_ix >= num_faces) {
    FMUS_EPRINTF(fmus->tw, "Error loading TrueType / OpenType font: Font does not contain requested "
                           "face %u.\n", desired_face_ix);
    e = FMUS_E_FT_FACE_NOT_FOUND;
  }
  
  goto Ldone;
  
Lfail:
  if (font->file_contents)
    FMUS_FREEMSG(font->file_contents, "fmus_read_file payload");
  font->file_contents = NULL;
  font->file_len = 0;
  if (font->faces)
    FMUS_FREEMSG(font->faces, "font->faces");
  font->faces = NULL;
Ldone:
  return e;
}

void fmus_text_finish(fmus_text_t *text) {
  u32_t j;
  fmus_font_t *font = NULL;
  font = &(text->font);
  for (j=0;j<font->num_faces;j++) {
    FT_Done_Face(font->faces[j]);
  }
  if (font->file_contents)
    FMUS_FREEMSG(font->file_contents, "fmus_read_file payload");
  font->file_contents = NULL;
  font->file_len = 0;
  if (font->faces)
    FMUS_FREEMSG(font->faces, "font->faces");
  font->faces = NULL;
  FT_Done_FreeType(text->library);
}
